#include <stdio.h>
#include <dir.h>
#include "renderer.h"
#include "config.h"

BOOL CALLBACK InitDlgProc(HWND,UINT,WPARAM,LPARAM);

Renderer render;

// 3D objects
FluidElement *fl;
LightElement *lt;
Object *mesh1;
Object *mesh2;
Object *fluid;
Object *light;
Starfield *sf;
Vector *atoms,*speed;
int nump;

// 2D objects
TexQuad2D *text1;
TexQuad2D *text2;
TexQuad2D *text3;
TexQuad2D *text4;

// Timing
float updateTime;
Timer clock;
float frameTime;
int actAnim=0;
float animTime=0;
float totTime=0;
#define ANIM 17
float anim[ANIM];

// Altro
bool scrollText=false;
int scroll=0;

void loadObjects() {
	
	anim[0]=14.0;  // Fade In
	anim[1]=10.0;  // Move Around
	anim[2]=7.0;   // Show Wireframe
	anim[3]=3.0;   // Show Atoms
	anim[4]=8.0;   // Move Drop
	anim[5]=14.0;  // Free Movement
	anim[6]=2.0;   // Fade Out
	anim[7]=2.0;   // Fade In
	anim[8]=1.0;   // Little Wait
	anim[9]=10.0;  // Gravity Drop w/text
	anim[10]=15.0; // Gravity Drop
	anim[11]=2.0;  // Fade Out
	anim[12]=2.0;  // Fade In
	anim[13]=8.0; // Stone Cup w/text
	anim[14]=40.0; // Stone Cup
	anim[15]=20.0; // Thanks
	anim[16]=4.0;  // Fade Out
	
	randomize();
	chdir("./data");
	
	// LIGHT ELEMENT
	light=new Object();
	lt=new LightElement();
	lt->setType(180,60.0,Vector(-1.0,-1.0,1.0),POSITIONAL);
	lt->setPosition(10.0,10.0,-10.0);
	lt->setColor(1.0,1.0,1.0);
	lt->setAttenuation(1.0,0.0,0.0);
	lt->enable(true,false,false);
	lt->visible(true);
	lt->setVisible(false);
	light->addElement(lt);
	render.addObject(light);

	// FLUID ELEMENT
	fluid=new Object();
	fl=new FluidElement();
	fl->create(7,22,1.5,1.5,1.5);
	fl->setAmbColor(0.6,0.6,0.6,1.0);
	fl->setDifColor(0.8,0.6,0.6,1.0);
	fl->setSpcColor(1.0,1.0,1.0,1.0);
	fl->setEmsColor(0.0,0.0,0.0,1.0);
	fl->setLevel(12.0);
	fl->setGravity(Vector(0.0,0.0,0.0));
	fl->setMass(0.01);
	fl->setDistance(0.4);
	fl->setViscosity(0.98);
	fl->setLimits(-100,100,-100,100,-100,100);
	fl->setEnvironmentMap("lava.bmp");
	fl->setPosition(0.0,0.0,0.0);
	fl->getAtoms(atoms,speed,nump);
	fluid->addElement(fl);
	fluid->system().heading(10.0);
	fluid->system().pitch(10.0);
	render.addObject(fluid);

	// MESH ELEMENT
	mesh1=new Object();
	load3DS("scene.3ds",mesh1);
	mesh1->setVisible(false);
	render.addObject(mesh1);
	mesh2=new Object();
	load3DS("stonecup.3ds",mesh2);
	mesh2->setVisible(false);
	render.addObject(mesh2);
	
	// STARFIELD BACKGROUND
	sf=new Starfield();
	sf->create(50.0,1.0,10000,"dummy.bmp",1.0);
	render.setBackground(sf);
	
	// TEXTURED QUAD 2D
	text1=new TexQuad2D;
	text1->create("text.bmp");
	text1->setSize(400,100);
	text1->setTexCoord(0,0.0,0.75);
	text1->setTexCoord(1,1.0,0.75);
	text1->setTexCoord(2,1.0,1.0);
	text1->setTexCoord(3,0.0,1.0);
	text1->setVisible(false);
	text2=new TexQuad2D;
	text2->setSize(280,70);
	text2->create("text.bmp");
	text2->setTexCoord(0,0.0,0.5);
	text2->setTexCoord(1,1.0,0.5);
	text2->setTexCoord(2,1.0,0.75);
	text2->setTexCoord(3,0.0,0.75);
	text2->setVisible(false);
	text3=new TexQuad2D;
	text3->setSize(600,150);
	text3->create("text.bmp");
	text3->setTexCoord(0,0.0,0.25);
	text3->setTexCoord(1,1.0,0.25);
	text3->setTexCoord(2,1.0,0.5);
	text3->setTexCoord(3,0.0,0.5);
	text3->setVisible(false);
	text4=new TexQuad2D;
	text4->setSize(280,70);
	text4->create("text.bmp");
	text4->setTexCoord(0,0.0,0.0);
	text4->setTexCoord(1,1.0,0.0);
	text4->setTexCoord(2,1.0,0.25);
	text4->setTexCoord(3,0.0,0.25);
	text4->setVisible(false);
	
	render.setAmbient(0.6f,0.6f,0.6f);
	render.setPerspective(45.0,0.5,1000.0);
	render.getCamera().move(0.0,0.0,-5.0);
	frameTime=1/30.0f;
	updateTime=0;
	render.setMaxFramerate(1.0f/frameTime,false);
	clock.init();
	clock.setTime();
	chdir("./..");
}

void killObjects() {
	//viene fatto da renderer	
}

void mainLoop() {
	if (fl->getVisible()) fl->updateAtoms(updateTime);
	
	
	// Keyframer
	float step,pro;
	step=(clock.getTime()/1000.0)-animTime; // step from prev frame
	animTime+=step;
	totTime+=step;
	
	// Transitions between animations
	if (animTime>anim[actAnim]) {
		animTime=0;
		actAnim++;
		clock.setTime();
		switch (actAnim) {
			case 1:  // Move Around
				render.setFadeQuad(0.0,0.0,0.0,0.0);
				updateTime=frameTime/2;
				break;
			case 3:  // Show Atoms
				fl->setMode(WIREFRAME);
				fl->visAtoms(true);
				break;
			case 4:  // Move Drop
				text2->setVisible(true);
				updateTime=0;
				break;
			case 5:  // Free Movement
				text1->setVisible(true);
				updateTime=frameTime/2;
				break;
			case 6:  // Fade Out
				text2->setVisible(false);
				break;
			case 7:  // Fade In
				render.getCamera().reset();
				render.getCamera().move(0.0,0.0,-5.0);
				mesh1->setVisible(true);
				text1->setVisible(false);
				fl->kill();
				fl->create(20,20,1.5,2.0,1.5);
				fl->setAmbColor(0.6,0.6,0.6,1.0);
				fl->setDifColor(0.6,0.6,0.6,1.0);
				fl->setSpcColor(1.0,1.0,1.0,1.0);
				fl->setEmsColor(0.0,0.0,0.0,1.0);
				fl->setLevel(36.0);
				fl->setGravity(Vector(0.0,0.0,0.0));
				fl->setMass(0.01);
				fl->setDistance(0.35);
				fl->setViscosity(0.94);
				fl->setLimits(-1.5,1.5,-1.1,100.0,-1.5,1.5);
				fl->setEnvironmentMap("./data/chrome.bmp");
				fl->setMode(SMOOTH);
				fl->visAtoms(false);
				fl->setPosition(0.0,0.0,0.0);
				fl->getAtoms(atoms,speed,nump);
				fl->setVisible(true);
				for(int i=0;i<3+nump/2;i++) {
					atoms[i].set(0.7-1.4*float(rand())/RAND_MAX,5.0-1.4*float(rand())/RAND_MAX,0.7-1.4*float(rand())/RAND_MAX);
					speed[i].set(0.0,0.0,0.0);
				}
				for(int i=3+nump/2;i<nump;i++) {
					atoms[i].set(0.7-1.4*float(rand())/RAND_MAX,15.0-1.4*float(rand())/RAND_MAX,0.7-1.4*float(rand())/RAND_MAX);
					speed[i].set(0.0,0.0,0.0);
				}
				atoms[0].set(0.25,1.5,0.25);
				atoms[1].set(0.0,1.5,0.45);
				atoms[2].set(-0.25,1.5,0.25);
				fluid->system().reset();
				updateTime=0;
				break;
			case 8:  // Little Wait
				render.setFadeQuad(0.0,0.0,0.0,0.0);
				break;
			case 9:  // Gravity Drop w/text
				updateTime=frameTime/1.7;
				fl->setGravity(Vector(0.0,-11.5,0.0));
				text3->setVisible(true);
				text3->setPosition(100,150);
				break;
			case 10:  // Gravity Drop
				text3->setVisible(false);
				break;
			case 12:  // Fade In
				render.getCamera().reset();
				render.getCamera().move(0.0,0.0,-8.0);
				mesh1->setVisible(false);
				mesh2->setVisible(true);
				mesh2->system().move(0.0,-3.0,0.0);
				fl->kill();
				fl->create(80,25,1.7,3.0,1.7);
				fl->setAmbColor(0.6,0.6,0.6,0.6);
				fl->setDifColor(0.4,0.4,0.8,0.6);
				fl->setSpcColor(1.0,1.0,1.0,0.6);
				fl->setEmsColor(0.0,0.0,0.0,0.6);
				fl->setLevel(34.0);
				fl->setGravity(Vector(0.0,-100.0,0.0));
				fl->setMass(0.02);
				fl->setDistance(6.0);
				fl->setViscosity(0.96);
				fl->setLimits(-1.1,1.1,-2.2,1000.0,-1.1,1.1);
				fl->setEnvironmentMap("./data/sky.bmp");
				fl->getAtoms(atoms,speed,nump);
				fl->setVisible(true);
				for(int i=0;i<2*nump/5;i++) {
					atoms[i].set(0.7-1.4*float(rand())/RAND_MAX,0.7-1.4*float(rand())/RAND_MAX,0.7-1.4*float(rand())/RAND_MAX);
					speed[i].set(0.0,0.0,0.0);
				}
				for(int i=2*nump/5;i<3*nump/5;i++) {
					atoms[i].set(0.1-0.2*float(rand())/RAND_MAX,10.0+50.0*float(rand())/RAND_MAX,0.1-0.2*float(rand())/RAND_MAX);
					speed[i].set(0.0,0.0,0.0);
				}
				for(int i=3*nump/5;i<nump;i++) {
					atoms[i].set(0.1-0.2*float(rand())/RAND_MAX,60.0+70.0*float(rand())/RAND_MAX,0.1-0.2*float(rand())/RAND_MAX);
					speed[i].set(0.0,0.0,0.0);
				}
				fluid->system().reset();
				updateTime=frameTime/4;
				break;
			case 13:  // Stone Cup w/text
				text4->setVisible(true);
				text4->setPosition(0.0,0.0);
				text4->setBrightness(0.0);
				render.setFadeQuad(0.0,0.0,0.0,0.0);
				break;
			case 14:  // Stone Cup
				text4->setVisible(false);
				break;
			case 15:  // Thanks
				scrollText=true;
				break;
			case ANIM: // Quit
				PostQuitMessage(0);
				break;
		}
	}

	// pro: progress of single animation form 0 to 1
	// step: duration of last step normalized to length
	pro=animTime/anim[actAnim];
	step=step/anim[actAnim];

	// Animations
	switch (actAnim) {
		case 0:  // Fade In
			render.setFadeQuad(0.0,0.0,0.0,1.0-pro);
			atoms[0].set(1.05,0.02,0.03);
			atoms[1].set(-1.06,0.01,0.03);
			atoms[2].set(0.04,1.05,0.04);
			atoms[3].set(0.02,-1.03,0.03);
			atoms[4].set(0.05,0.02,1.08);
			atoms[5].set(0.01,0.03,-1.01);
			atoms[6].set(0.0,0.0,0.0);
			break;
		case 1:  // Move around
			render.getCamera().reset();
			render.getCamera().heading(pro*360);
			render.getCamera().roll(pro*360);
			render.getCamera().move(0.0,0.0,-5.0);
			break;
		case 2:  // Show WireFrame
			if (rand()%3!=0) fl->setMode(WIREFRAME);
			else fl->setMode(SMOOTH);
			break;
		case 3:  // Show Atoms
			break;
		case 4:  // Move Drop
			atoms[6].set(pro*1.1,pro*1.1,-pro*1.1);
			text2->setPosition(-200+500*pro,20);
			text2->setBrightness(2*pro);
			break;
		case 5:  // Free Movement
			text1->setPosition(400,500-490*pro);
			text1->setBrightness(pro*2);
			text2->setBrightness(1.0-3*pro);
			break;
		case 6:  // Fade Out
			render.setFadeQuad(0.0,0.0,0.0,pro);
			break;
		case 7:  // Fade In
			render.setFadeQuad(0.0,0.0,0.0,1.0-pro);
			atoms[0].set(0.3,1.5,0.3);
			atoms[1].set(0.0,1.5,0.5);
			atoms[2].set(-0.3,1.5,0.3);
			break;
		case 9:  // Gravity Drop w/text
			atoms[0].set(0.3,1.5,0.3);
			atoms[1].set(0.0,1.5,0.5);
			atoms[2].set(-0.3,1.5,0.3);
			render.getCamera().reset();
			render.getCamera().heading(2*pro*360);
			render.getCamera().move(0.0,0.0,-5.0);
			text3->setBrightness(sin(3.14*pro)*sin(3.14*pro));
			break;
		case 10:  // Gravity Drop
			atoms[0].set(0.3,1.5,0.3);
			atoms[1].set(0.0,1.5,0.5);
			atoms[2].set(-0.3,1.5,0.3);
			render.getCamera().reset();
			render.getCamera().heading(4*pro*360);
			render.getCamera().move(0.0,0.0,-5.0);
			break;
		case 11:  // Fade Out
			render.setFadeQuad(0.0,0.0,0.0,pro);
			break;
		case 12:  // Fade In
			render.setFadeQuad(0.0,0.0,0.0,1.0-pro);
			break;
		case 13:  // Stone Cup w/text
			render.getCamera().reset();
			render.getCamera().pitch(30*pro);
			render.getCamera().move(0.0,0.0,-8.0);
			text4->setPosition(500*pro,525*pro);
			text4->setBrightness(sin(3.14*pro)*sin(3.14*pro));
			break;
		case 14:  // Stone Cup
			render.getCamera().reset();
			render.getCamera().heading(4*pro*360);
			render.getCamera().pitch(30);
			render.getCamera().move(0.0,0.0,-8.0);
			break;
		case 15:  // Thanks
			render.getCamera().reset();
			render.getCamera().heading(2*pro*360);
			render.getCamera().pitch(30);
			render.getCamera().move(0.0,0.0,-8.0);
			scroll=500*pro;
			break;
		case 16:  // Fade Out
			render.setFadeQuad(0.0,0.0,0.0,pro);
			break;
	}	
}

void printScr() {
	text1->render();
	text2->render();
	text3->render();
	text4->render();
	if (scrollText) {
		render.getFont().print(40,-20+scroll,"AquaDemo - GameDev Physics Contest");
		render.getFont().print(40,-40+scroll,"By Fuzzware 2002");
		render.getFont().print(40,-100+scroll,"Features fluid simulation by means of particle systems and");
		render.getFont().print(40,-120+scroll,"scalar field polygonisation with the marching cubes algorithm");
		render.getFont().print(40,-140+scroll,"Additional: starfield and 3ds meshes");
		render.getFont().print(40,-200+scroll,"Thanks to all the people that helped me in various ways (testing,");
		render.getFont().print(40,-220+scroll,"discussions), explicitly to Paul Bourke for his online article");
		render.getFont().print(40,-240+scroll,"about marching cubes algorithm.");
		render.getFont().print(40,-300+scroll,"Bug report, comments and other:");
		render.getFont().print(40,-320+scroll,"teofuzz@inwind.it");
		render.getFont().print(40,-380+scroll,"See you again...");
	}
}

int WINAPI WinMain(HINSTANCE hInstance,HINSTANCE hPrevInstance,LPSTR lpCmdLine,int nCmdShow) {
	hPrevInstance,lpCmdLine,nCmdShow;
	openMsg();
	char title[]="AquaDemo by Fuzzware2002";
	SettingsInfo sInfo;
	bool cfgRes;
	do {
		cfgRes=displayConfigDialog(hInstance,sInfo);
		if (!cfgRes) {
			closeMsg();
			return 0;
		}
		disp.setMode(sInfo.wscr,sInfo.hscr,sInfo.bpp,sInfo.full,sInfo.vsync,title);
		message("Selected display: %dx%dx%d ",sInfo.wscr,sInfo.hscr,sInfo.bpp);
		if (sInfo.full) message("(fullscreen,");
		else message("(windowed,");
		if (sInfo.vsync) message("vsync OFF)\n");
		else message("vsync ON)\n");
	} while(!disp.enableOpenGL());
	render.init(disp.getHDC());
	
	loadObjects();
	snd.playSong("./data/amns.mp3");
	render.setCallBack(MAINLOOP,mainLoop);
	render.setCallBack(DRAW2D,printScr);
	message("OpenGL version: ");
	message((char*)glGetString(GL_VERSION));
	message("\nRenderer: ");
	message((char*)glGetString(GL_RENDERER));
	message("\nVendor: ");
	message((char*)glGetString(GL_VENDOR));
	message("\nExtensions: ");
	message((char*)glGetString(GL_EXTENSIONS));
	message("\n\n");
	render.mainLoop();
	killObjects();
	render.unInit();
	disp.disableOpenGL();
	closeMsg();
	return 0;
}

