// *****************************
// ***     D r i v e r s     ***
// *****************************
// ***    by Fuzzware 2001   ***
// *****************************
// 
// Inizializzazione del display,
// del suono

#include "drivers.h"

BOOL (APIENTRY *wglSwapIntervalEXT)(int);
Display disp;
Keyboard kbd;
Sound snd;

LRESULT CALLBACK WndProc(HWND hWnd,UINT msg,WPARAM wParam,LPARAM lParam) {
	bool dn;
	int sc;
	switch (msg) {
		case WM_SYSCOMMAND:
			switch (wParam) {
				case SC_SCREENSAVE:
				case SC_MONITORPOWER:
				return 0;
			}
			break;
		case WM_CREATE:
			return 0;
		case WM_CLOSE:
			PostQuitMessage(0);
			return 0;
		case WM_DESTROY:
			return 0;
		case WM_CHAR:
			return 0;
		case WM_KEYDOWN:
		case WM_KEYUP:
			dn=(msg==WM_KEYDOWN);
			sc=(lParam>>16)&0xff;
			switch(wParam) {
				case VK_ESCAPE:
					PostQuitMessage(0);
					return 0;
				case VK_UP:
      			kbd.keyDown(k_UP,dn);
      			return 0;
      		case VK_DOWN:
      			kbd.keyDown(k_DOWN,dn);
   	   		return 0;
      		case VK_LEFT:
      			kbd.keyDown(k_LEFT,dn);
      			return 0;
      		case VK_RIGHT:
      			kbd.keyDown(k_RIGHT,dn);
	      		return 0;
	      	case VK_F1:
      			kbd.keyDown(k_F1,dn);
	      		return 0;
	      	case VK_F2:
      			kbd.keyDown(k_F2,dn);
	      		return 0;
	      	case VK_F3:
      			kbd.keyDown(k_F3,dn);
	      		return 0;
	      	case VK_F4:
      			kbd.keyDown(k_F4,dn);
	      		return 0;
	      	case VK_F5:
      			kbd.keyDown(k_F5,dn);
	      		return 0;
	      	case VK_F6:
      			kbd.keyDown(k_F6,dn);
	      		return 0;
	      	case VK_F7:
      			kbd.keyDown(k_F7,dn);
	      		return 0;
	      	case VK_F8:
      			kbd.keyDown(k_F8,dn);
	      		return 0;
	      	case VK_F9:
      			kbd.keyDown(k_F9,dn);
	      		return 0;
	      	case VK_F10:
      			kbd.keyDown(k_F10,dn);
	      		return 0;
	      	case VK_F11:
      			kbd.keyDown(k_F11,dn);
	      		return 0;
	      	case VK_F12:
      			kbd.keyDown(k_F12,dn);
	      		return 0;
   	   }
   	   if (sc<k_LAST) kbd.keyDown(sc,dn);
      	return 0;
		case WM_SIZE:
			return 0;
		default:
  			return DefWindowProc(hWnd,msg,wParam,lParam);
	}
	return 0;
}


// *****************************
// *****************************
// ***     D i s p l a y     ***
// *****************************
// *****************************

Display::Display() {
	wscr=800;
	hscr=600;
	bpp=16;
	full=true;
	vsync=true;
	title="OpenGL";
	winHandle=NULL;
	devHandle=NULL;
	renHandle=NULL;
	hInst=GetModuleHandle(NULL);
}

Display::~Display() {
	disableOpenGL();
}

void Display::setMode(int w,int h,int b,bool f,bool v,char *t) {
	if (winHandle!=NULL) return;
	wscr=w;
	hscr=h;
	bpp=b;
	full=f;
	vsync=v;
	title=t;
}

bool Display::enableOpenGL() {
	WNDCLASS wc;
	wc.style=CS_HREDRAW|CS_VREDRAW|CS_OWNDC;
	wc.lpfnWndProc=(WNDPROC)WndProc;
	wc.cbClsExtra=0;
	wc.cbWndExtra=0;
	wc.hInstance=hInst;
	wc.hIcon=LoadIcon(NULL,IDI_WINLOGO);
	wc.hCursor=LoadCursor(NULL,IDC_ARROW);
	wc.hbrBackground=NULL;
	wc.lpszMenuName=NULL;
	wc.lpszClassName=WNDCLSNAME;
	message("Registering window class... ");
	if (!RegisterClass(&wc)) {
		message("failed\n\n");
		return false;
	}
	else message ("ok\n");
	if (full) {
		message("Setting fullscreen mode... ");
		DEVMODE dmScreenSettings;
		memset(&dmScreenSettings,0,sizeof(dmScreenSettings));
		dmScreenSettings.dmSize=sizeof(dmScreenSettings);
		dmScreenSettings.dmPelsWidth=wscr;
		dmScreenSettings.dmPelsHeight=hscr;
		dmScreenSettings.dmBitsPerPel=bpp;
		dmScreenSettings.dmFields=DM_BITSPERPEL|DM_PELSWIDTH|DM_PELSHEIGHT;
		if (ChangeDisplaySettings(&dmScreenSettings,CDS_FULLSCREEN)!=DISP_CHANGE_SUCCESSFUL) {
			message("failed.\n");
			if (MessageBox(NULL,"The Requested Fullscreen Mode Is Not Supported By\nYour Video Card. Use Windowed Mode Instead?","Question",MB_YESNO|MB_ICONEXCLAMATION)==IDYES) {
				full=false;
				message("Using windowed mode.\n");
			}
			else {
				message("Quitting.\n\n");
				return false;
			}
		}
		else message("ok.\n");
	}
	DWORD dwExStyle;
	DWORD dwStyle;
	RECT WindowRect;
	WindowRect.left=(long)0;
	WindowRect.right=(long)wscr;
	WindowRect.top=(long)0;
	WindowRect.bottom=(long)hscr;
	if (full) {
		dwExStyle=WS_EX_APPWINDOW;dwStyle=WS_POPUP;
		ShowCursor(false);
	}
	else {
		dwExStyle=WS_EX_APPWINDOW|WS_EX_WINDOWEDGE;
		dwStyle=WS_OVERLAPPEDWINDOW;
	}
	AdjustWindowRectEx(&WindowRect,dwStyle,FALSE,dwExStyle);
	message("Creating window... ");
	winHandle=CreateWindowEx(dwExStyle,WNDCLSNAME,title,
									WS_CLIPSIBLINGS|WS_CLIPCHILDREN|dwStyle,0,0,
									WindowRect.right-WindowRect.left,
									WindowRect.bottom-WindowRect.top,
									NULL,NULL,hInst,NULL);
	if (!winHandle) {
		message("failed.\n\n");
		disableOpenGL();
		return false;
	}
	message("ok.\n");
	int pixelFormat;
	static PIXELFORMATDESCRIPTOR pfd;
	memset(&pfd,0,sizeof(PIXELFORMATDESCRIPTOR));
	pfd.nSize=sizeof(pfd);
	pfd.nVersion=1;
	pfd.dwFlags=PFD_DRAW_TO_WINDOW|PFD_SUPPORT_OPENGL|PFD_DOUBLEBUFFER;
	pfd.iPixelType=PFD_TYPE_RGBA;
	pfd.cColorBits=(BYTE)bpp;
	pfd.cDepthBits=16;
	pfd.iLayerType=PFD_MAIN_PLANE;
	message("Creating a device context... ");
	devHandle=GetDC(winHandle);
	if (!devHandle) {
		message("failed.\n\n");
		disableOpenGL();
		return false;
	}
	message("ok.\n");
	message("Choosing pixel format... ");
	pixelFormat=ChoosePixelFormat(devHandle,&pfd);
	if (!pixelFormat)	{
		message("failed.\n\n");
		disableOpenGL();
		return false;
	}
	message("ok.\n");
	DescribePixelFormat(devHandle,pixelFormat,sizeof(PIXELFORMATDESCRIPTOR),&pfd); 
	message("\nPixel format description\n");
	message("Color buffer: %d\n",pfd.cColorBits);
	message("Depth buffer: %d\n",pfd.cDepthBits);
	message("Stencil buffer: %d\n",pfd.cStencilBits);
	message("Accumulation buffer: %d\n\n",pfd.cAccumBits);
	if (((pfd.dwFlags&PFD_GENERIC_ACCELERATED)==0)&&((pfd.dwFlags&PFD_GENERIC_FORMAT)!=0)) {
		message("No hardware acceleration found.\n");
		disableOpenGL();
		MessageBox(NULL,"No hardware acceleration found\nwith this resolution.","Info",MB_OK|MB_ICONEXCLAMATION);
		return false;
	}
	message("Setting pixel format... ");
	if (!SetPixelFormat(devHandle,pixelFormat,&pfd)) {
		message("failed.\n\n");
		disableOpenGL();
		return false;
	}
	message("ok.\n");
	message("Creating rendering context... ");
	renHandle=wglCreateContext(devHandle);
	if (!renHandle) {
		message("failed.\n\n");
		disableOpenGL();
		return false;
	}
	message("ok.\n");
	message("Activating rendering context... ");
	if (!wglMakeCurrent(devHandle,renHandle)) {
		message("failed.\n\n");
		disableOpenGL();
		return false;
	}
	message("ok.\n");
	if (vsync) {
		wglSwapIntervalEXT=(BOOL(APIENTRY*)(int))wglGetProcAddress("wglSwapIntervalEXT");
		if (wglSwapIntervalEXT!=NULL) {
			message("Vsync disabled (wglSwapIntervalEXT)\n");
			wglSwapIntervalEXT(0);
		}
		else {
			message("Vsync unchanged (wglSwapIntervalEXT not found)\n");
		}
	}
	ShowWindow(winHandle,SW_SHOW);
	SetForegroundWindow(winHandle);
	SetFocus(winHandle);
	resizeScene(wscr,hscr);
	message("\n-- OpenGL display ENABLED --\n\n");
	return true;
}


void Display::disableOpenGL() {
	if (full)	{
		ChangeDisplaySettings(NULL,0);
		ShowCursor(TRUE);
	}	
	if (renHandle) {
		message("Deactivating rendering context... ");
		if (!wglMakeCurrent(NULL,NULL)) message("failed.\n");
		else message("ok.\n");
		message("Releasing rendering context... ");
		if (!wglDeleteContext(renHandle)) message("failed.\n");
		else message("ok.\n");
		renHandle=NULL;
	}
	if (devHandle) {
		message("Releasing device context... ");
		if (!ReleaseDC(winHandle,devHandle)) message("failed.\n");
		else message("ok.\n");
		devHandle=NULL;
	}
	if (winHandle) {
		message("Destroying window... ");
		if (!DestroyWindow(winHandle)) message("failed.\n");
		else message("ok.\n");
		winHandle=NULL;
	}
	if (UnregisterClass(WNDCLSNAME,hInst)) message("Unregistering window class... ok.\n");
	if (hInst) message("\n-- OpenGL display DISABLED --\n\n");
	hInst=NULL;
}

void Display::resizeScene(int w,int h) {
	wscr=w;hscr=h;
	if (hscr==0) hscr=1;
	glViewport(0,0,wscr,hscr);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60.0,(GLfloat)wscr/(GLfloat)hscr,0.5,200.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

void Display::screenShot() {
	unsigned char *buf;
	buf=new unsigned char[wscr*hscr*3];
	glReadPixels(0,0,wscr,hscr,GL_RGB,GL_UNSIGNED_BYTE,buf);
	Image img;
	img.setImage(buf,wscr,hscr);
	img.saveBMP("screenshot.bmp");
}

int Display::getWidth() {
	return wscr;
}

int Display::getHeight() {
	return hscr;
}

HDC Display::getHDC() {
	return devHandle;
}


// *****************************
// *****************************
// ***    K e y b o a r d    ***
// *****************************
// *****************************

Keyboard::Keyboard() {
	for(int i=0;i<k_LAST;i++)
		Keyboard::keys[i]=Keyboard::push[i]=false;
}

Keyboard::~Keyboard() {
}

void Keyboard::keyDown(int k,bool dn) {
	if ((!keys[k])&&(dn)) push[k]=true;
	keys[k]=dn;
}

bool Keyboard::readKey(int k) {
	return keys[k];
}

bool Keyboard::readPush(int k) {
	bool tmp=push[k];
	push[k]=false;
	return tmp;
}

// *****************************
// *****************************
// ***       S o u n d       ***
// *****************************
// *****************************

Sound::Sound() {
	FSOUND_SetOutput(-1);
	FSOUND_SetDriver(0);
	FSOUND_Init(44100,16,0);
}

Sound::~Sound() {
	FSOUND_Close();
}

void Sound::playSong(char *fname) {
	FSOUND_STREAM *bsong=FSOUND_Stream_OpenFile(fname,FSOUND_LOOP_NORMAL,0);
	FSOUND_Stream_Play(FSOUND_FREE,bsong);
}