// *****************************
// ***      L o a d e r      ***
// *****************************
// ***    by Fuzzware 2001   ***
// *****************************
// 
// Caricatore di file 3d studio
// con vertici e materiali, e land

#include "loader.h"


// *****************************
// ***   3 d s l o a d e r   ***
// *****************************

char objName[50];
char *matName;
ElementPtr *elem;
char texName[256];
int editEnd,matsEnd,objsEnd;
unsigned short nv,nf,nm;
int num;

ChunkHeader getChunk(FILE *in) {
	ChunkHeader res;
	fread(&(res.ID),2,1,in);
	fread(&(res.length),4,1,in);
	return res;
}

bool load3DS(char *fname,Object *obj) {
	FILE *in;
	ChunkHeader block;
	if (fname==NULL) return false;
	message("Loading 3ds file: ",fname,"\n");
	if ((in=fopen(fname,"rb"))==NULL) {
		message("Error: could not open file!\n\n");
		return false;
	}
	block=getChunk(in);
	if (block.ID!=MAIN3DS) {
		fclose(in);
		message("Error: not a valid 3ds format!\n\n");
 		return false;
	}
	getObjNumber(in);
	message("Found %d objects.\n",num);
	message("Creating %d elements.\n",num);
	elem=new ElementPtr[num];
	for(int i=0;i<num;i++) elem[i]=new MeshElement();
	matName=new char[num*50];
	for(int i=0;i<num*50;i++) matName[i]=0;
	fseek(in,0,SEEK_SET);
	block=getChunk(in);
	message("\nStart reading MESH\n");
	indent();
	readMesh(in);
	deindent();
	message("End reading MESH\n");
	fseek(in,0,SEEK_SET);
	block=getChunk(in);
	message("Start reading MATERIALS\n");
	indent();
	readMaterials(in);
	deindent();
	message("End reading MATERIALS\n");
	fclose(in);
	message("Closing 3ds file.\n\n");
	for(int i=0;i<num;i++)
		obj->addElement(elem[i]);
	delete[] matName;
	delete[] elem;
	return true;
}

void getObjNumber(FILE *in) {
	ChunkHeader block;
	int i;
	float f;
	num=0;
	char tmp[50];
	int filelen=filelength(fileno(in));
	while(!feof(in)&&ftell(in)<filelen) { //First while
		block=getChunk(in);
		switch(block.ID) { //First switch
			case EDIT3DS:
				editEnd=ftell(in)+(block.length-6);
				while (!feof(in)&&ftell(in)<editEnd) { //Second while
					block=getChunk(in);		
					switch(block.ID) { //Second switch
						case MATERIALS:
							matsEnd=ftell(in)+(block.length-6);
							while (!feof(in)&&ftell(in)<matsEnd) { //Materials while
								block=getChunk(in);
								switch(block.ID) { //Materials switch
									case MATERIALNAME:
										i=0;
										do {
											fread(&tmp[i++],1,1,in);
										} while(tmp[i-1]!='\0');
										//nMat++; se prima o poi uso tutti i materiali...
										break;
									default:
										fseek(in,block.length-6,SEEK_CUR);
								} //Materials switch
							} //Materials while
							break;
						case OBJECT3DS:
							objsEnd=ftell(in)+(block.length-6);
							i=0;
							do {
								fread(&objName[i++],1,1,in);
							} while(objName[i-1]!='\0');
							while(!feof(in)&&ftell(in)<objsEnd) { //Object while
								block=getChunk(in);
								switch(block.ID) { //Object switch
									case POLYLIST:
										break;
									case VERTEXLIST:
										fread(&nv,sizeof(unsigned short),1,in);
										for(i=0;i<nv;i++) {
											fread(&f,sizeof(float),1,in);
											fread(&f,sizeof(float),1,in);
											fread(&f,sizeof(float),1,in);
										}
										break;
									/*case OBJECT3DS:
										fseek(in,-6,SEEK_CUR);
										done=true;
										break;*/
									default:
										fseek(in,block.length-6,SEEK_CUR);
								} //Object switch
							} //Object while
							if (nv!=0) num++;
							break;
						default:
							fseek(in,block.length-6,SEEK_CUR);
					} //Second switch
				} //Second while
				break;
			default:
				fseek(in,block.length-6,SEEK_CUR);
		} //First switch
	} //First while
	return;
}

void readMesh(FILE *in) {
	ChunkHeader block;
	Vector *vert=0;
	MapCoord *mapc=0;
	Triangle *face;
	unsigned short faceInfo;
	int i;
	int currObj=-1;
	int filelen=filelength(fileno(in));
	while(!feof(in)&&ftell(in)<filelen) { //First while
		block=getChunk(in);
		switch(block.ID) { //First switch
			case EDIT3DS:
				editEnd=ftell(in)+(block.length-6);
				while (!feof(in)&&ftell(in)<editEnd) { //Second while
					block=getChunk(in);		
					switch(block.ID) { //Second switch
						case OBJECT3DS:
							objsEnd=ftell(in)+(block.length-6);
							i=0;
							do {
								fread(&objName[i++],1,1,in);
							} while(objName[i-1]!='\0');
							message("Reading object: %s\n",objName);
							while(!feof(in)&&ftell(in)<objsEnd) { //Object while
								block=getChunk(in);
								switch(block.ID) { //Object switch
									case POLYLIST:
										break;
									case VERTEXLIST:
										fread(&nv,sizeof(unsigned short),1,in);
										message("Vertices: %d\n",nv);
										vert=new Vector[nv];
										for(i=0;i<nv;i++) {
											fread(&vert[i].x,sizeof(float),1,in);
											fread(&vert[i].y,sizeof(float),1,in);
											fread(&vert[i].z,sizeof(float),1,in);
										}
										break;
									case MAPPINGLIST:
										fread(&nm,sizeof(unsigned short),1,in);
										message("Mapping: %d\n",nm);
										mapc=new MapCoord[nm];
										for(i=0;i<nm;i++) {
											fread(&mapc[i].u,sizeof(float),1,in);
											fread(&mapc[i].v,sizeof(float),1,in);
										}
										break;
									case FACELIST:
										fread(&nf,sizeof(unsigned short),1,in);
										message("Faces: %d\n",nf);
										face=new Triangle[nf];
										for(i=0;i<nf;i++) {
											fread(&face[i].a,sizeof(unsigned short),1,in);
											fread(&face[i].b,sizeof(unsigned short),1,in);
											fread(&face[i].c,sizeof(unsigned short),1,in);
											fread(&faceInfo,sizeof(unsigned short),1,in);
										}
										currObj++;
										((MeshElement*)elem[currObj])->create(nv,nf,vert,mapc,face);
										vert=0;mapc=0;
										break;
									case MATERIALLIST:
										i=0;
										do {
											fread(&matName[50*currObj+i],1,1,in);
											i++;
										} while(matName[50*currObj+i-1]!='\0');
										message("Material: %s\n",&matName[50*currObj]);
										fread(&nf,sizeof(unsigned short),1,in);
										for(i=0;i<nf;i++) {
											fread(&faceInfo,sizeof(unsigned short),1,in);
										}
										break;
									default:
										fseek(in,block.length-6,SEEK_CUR);
								} //Object switch
							} //Object while
							break;
						default:
							fseek(in,block.length-6,SEEK_CUR);
					} //Second switch
				} //Second while
				break;
			default:
				fseek(in,block.length-6,SEEK_CUR);
		} //First switch
	} //First while
	return;
}

void readMaterials(FILE *in) {
	ChunkHeader block;
	int i;
	int filelen=filelength(fileno(in));
	int currObj=-1;
	char tmp[50];
	Material mat;
	bool amb,dif,spc;
	amb=dif=spc=false;
	float red,green,blue;
	while(!feof(in)&&ftell(in)<filelen) { //First while
		block=getChunk(in);
		switch(block.ID) { //First switch
			case EDIT3DS:
				editEnd=ftell(in)+(block.length-6);
				while (!feof(in)&&ftell(in)<editEnd) { //Second while
					block=getChunk(in);		
					switch(block.ID) { //Second switch
						case MATERIALS:
							matsEnd=ftell(in)+(block.length-6);
							while (!feof(in)&&ftell(in)<matsEnd) { //Materials while
								block=getChunk(in);
								switch(block.ID) { //Materials switch
									case MATERIALNAME:
										i=0;
										do {
											fread(&tmp[i++],1,1,in);
										} while(tmp[i-1]!='\0');
										message("Reading material: %s\n",matName);
										currObj=-1;
										amb=dif=spc=false;
										for(i=0;i<num;i++)
											if (strcmp(tmp,&matName[i*50])==0) currObj=i;
										break;
									case MATAMBIENT:
										message("Ambient: ");
										amb=true;dif=spc=false;
										break;
									case MATDIFFUSE:
										message("Diffuse: ");
										dif=true;amb=spc=false;
										break;
									case MATSPECULAR:
										message("Specular: ");
										spc=true;amb=dif=false;
										break;
									case MATTEXTURE:
										//tex=true;
										break;
									case MATTEXNAME:
										i=0;
										do {
											fread(&texName[i++],1,1,in);
										} while(texName[i-1]!='\0');
										message("Texture: %s\n",texName);
										for(i=0;i<num;i++)
											if (strcmp(tmp,&matName[i*50])==0) ((MeshElement*)elem[i])->setDiffuseMap(texName);
										break;
									case MATCOLOR:
										unsigned char r, g, b;
										fread(&r,1,1,in); // Red component 1 byte
										fread(&g,1,1,in); // Green component 1 byte
										fread(&b,1,1,in); // Blue component 1 byte
										message("R=%d G=%d B=%d\n",r,g,b);
										if (currObj>=0) {
											red=(float)r/255.0;
											green=(float)g/255.0;
											blue=(float)b/255.0;
											if (amb) mat.setAmbient(red,green,blue,1.0);
											if (dif) mat.setDiffuse(red,green,blue,1.0);
											if (spc) mat.setSpecular(red,green,blue,1.0);
										}
										for(i=0;i<num;i++)
											if (strcmp(tmp,&matName[i*50])==0) ((MeshElement*)elem[i])->setMaterial(mat);
										break;
									default:
										fseek(in,block.length-6,SEEK_CUR);
								} //Materials switch
							} //Materials while
							break;
						default:
							fseek(in,block.length-6,SEEK_CUR);
					} //Second switch
				} //Second while
				break;
			default:
				fseek(in,block.length-6,SEEK_CUR);
		} //First switch
	} //First while
	return;
}

