// *****************************
// ***   M a t e r i a l s   ***
// *****************************
// ***    by Fuzzware 2001   ***
// *****************************
// 
// Incapsulamento delle propriet
// dei materiali

#include "materials.h"

Color::Color() {
	r=g=b=a=1.0;
}

void Color::setColor(float rr,float gg,float bb,float aa) {
	r=rr;g=gg;b=bb;a=aa;
}

Material::Material() {
	amb.setColor(0.2,0.2,0.2,1.0);
	dif.setColor(0.8,0.8,0.8,1.0);
	spc.setColor(0.0,0.0,0.0,1.0);
	ems.setColor(0.0,0.0,0.0,1.0);
	shine=10.0;// Attn: valore basso di shine sfuma male
	bump=env=-1.0;
	opaque=true;
}

Material::~Material() {
	kill();
}

void Material::setAmbient(float rr,float gg,float bb,float aa) {
	amb.setColor(rr,gg,bb,aa);
}

void Material::setDiffuse(float rr,float gg,float bb,float aa) {
	dif.setColor(rr,gg,bb,aa);
}

void Material::setSpecular(float rr,float gg,float bb,float aa) {
	spc.setColor(rr,gg,bb,aa);
}

void Material::setEmissive(float rr,float gg,float bb,float aa) {
	ems.setColor(rr,gg,bb,aa);
}

void Material::setShininess(float n) {
	shine=n;
}

void Material::setTransparency(float alpha) {
	amb.a=dif.a=spc.a=alpha;
}

bool Material::setDiffuseMap(char *fname,int detail) {
	return diffMap.setDiffuseMap(fname,detail);
}

bool Material::setBumpMap(char *fname,int detail) {
	bumpMap1.setBumpMap(fname,detail,BUMP1);
	return bumpMap2.setBumpMap(fname,detail,BUMP2);
}

bool Material::setEnvironmentMap(char *fname,int detail) {
	return envMap.setDiffuseMap(fname,detail);
}

bool Material::setGlossMap(char *fname,int detail) {
	return glossMap.setDiffuseMap(fname,detail);
}

void Material::applyColor() {
	glMaterialf(GL_FRONT_AND_BACK,GL_SHININESS,shine);
	GLfloat tmp1[]={amb.r,amb.g,amb.b,amb.a};
	GLfloat tmp2[]={dif.r,dif.g,dif.b,dif.a};
	GLfloat tmp3[]={spc.r,spc.g,spc.b,spc.a};
	GLfloat tmp4[]={ems.r,ems.g,ems.b,ems.a};
	glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT,tmp1);
	glMaterialfv(GL_FRONT_AND_BACK,GL_DIFFUSE,tmp2);
	glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,tmp3);
	glMaterialfv(GL_FRONT_AND_BACK,GL_EMISSION,tmp4);
}

void Material::applyDiffuse() {
	diffMap.apply();
}

void Material::applyBump1() {
	bumpMap1.apply();
}

void Material::applyBump2() {
	bumpMap2.apply();
}

void Material::applyEnvironment() {
	GLfloat tmp1[]={1.0,1.0,1.0,env};
	GLfloat tmp2[]={1.0,1.0,1.0,env};
	GLfloat tmp3[]={1.0,1.0,1.0,env};
	GLfloat tmp4[]={1.0,1.0,1.0,env};
	glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT,tmp1);
	glMaterialfv(GL_FRONT_AND_BACK,GL_DIFFUSE,tmp2);
	glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,tmp3);
	glMaterialfv(GL_FRONT_AND_BACK,GL_EMISSION,tmp4);
	envMap.apply();
}

void Material::applyGloss() {
	glossMap.apply();
}

void Material::kill() {
	diffMap.kill();
	bumpMap1.kill();
	bumpMap2.kill();
	envMap.kill();
	glossMap.kill();
}
