// *****************************
// ***     O b j e c t s     ***
// *****************************
// ***    by Fuzzware 2001   ***
// *****************************
// 
// Oggetto generico e conseguenti
// specializzazioni

#include "objects.h"

PFNGLMULTITEXCOORD1FARBPROC glMultiTexCoord1fARB=NULL;
PFNGLMULTITEXCOORD2FARBPROC glMultiTexCoord2fARB=NULL;
PFNGLMULTITEXCOORD3FARBPROC glMultiTexCoord3fARB=NULL;
PFNGLMULTITEXCOORD4FARBPROC glMultiTexCoord4fARB=NULL;
PFNGLACTIVETEXTUREARBPROC glActiveTextureARB=NULL;
PFNGLCLIENTACTIVETEXTUREARBPROC glClientActiveTextureARB=NULL;
bool multiTex=false;
int texUnits=1;
GlobalInfo gInfo;

// *******************************************
// *******************************************
// ****    3  D    O  B  J  E  C  T  S    ****
// *******************************************
// *******************************************



// *****************************
// *****************************
// ***     E l e m e n t     ***
// *****************************
// *****************************

Element::Element() {
	posx=posy=posz=0.0;
	center=Vector(0.0,0.0,0.0);
	antialias=false;
	visible=true;
	mode=SMOOTH;
}

Element::~Element() {
}

void Element::antialiasing(bool aa) {
	antialias=aa;
}

void Element::setMode(int m) {
	mode=m;
}

void Element::setVisible(bool b) {
	visible=b;
}

bool Element::getVisible() {
	return visible;
}

void Element::setInfo(LocalInfo *li) {
	info=li;
}

// *****************************
// *****************************
// ***      O b j e c t      ***
// *****************************
// *****************************

Object::Object() {
	num=0;
}

Object::~Object() {
	kill();
}

void Object::kill() {
	for(int i=0;i<num;i++)
		delete (elem[i]);
	if (num!=0) delete[] elem;
	num=0;
}

void Object::addElement(Element *el) {
	ElementPtr *tmp=new ElementPtr[num+1];
	for(int i=0;i<num;i++)
		tmp[i]=elem[i];
	tmp[num]=el;
	el->setInfo(&info);
	if (num!=0) delete[] elem;
	num++;
	elem=tmp;
}

void Object::transform() {
	ref.apply();
	// Matrice dell'oggetto
	for(int i=0;i<16;i++)
		info.objectMat[i]=ref.mat[i];
	// Posizione oggetto
	info.objectPos=ref.getEyePoint();	
}

void Object::localLight() {
	// Luce in coordinate oggetto
	float *m=ref.mat;
	Vector xv(m[0],m[4],m[8]);
	Vector yv(m[1],m[5],m[9]);
	Vector zv(m[2],m[6],m[10]);
	Matrix trasf(xv,yv,zv);
	Vector lt=gInfo.lightPos;
	lt.x=m[12]-lt.x;
	lt.y=m[13]-lt.y;
	lt.z=m[14]-lt.z;
	lt=trasf.xForm(lt);
	info.lightPosObj=lt;
}

void Object::render() {
	for(int i=0;i<num;i++)
		elem[i]->render();
}

void Object::antialiasing(bool aa) {
	for(int i=0;i<num;i++)
		elem[i]->antialiasing(aa);
}

void Object::setMode(int m) {
	for(int i=0;i<num;i++)
		elem[i]->setMode(m);
}

void Object::setVisible(bool b) {
	for(int i=0;i<num;i++)
		elem[i]->setVisible(b);
}

xyzSystem& Object::system() {
	return ref;
}

Element* Object::getElement(int n) {
	if ((n>=0)&&(n<num))	return elem[n];
	else return 0;
}

int Object::getNum() {
	return num;
}


// *****************************
// *****************************
// ***       W o r l d       ***
// *****************************
// *****************************

World::World() {
	num=0;
	numTot=0;
	back=0;
}

World::~World() {
	kill();
}

void World::kill() {
	for(int i=0;i<num;i++)
		delete (obj[i]);
	if (num!=0) delete[] obj;
	if (back!=0) delete back;
	if (numTot!=0) {
		delete[] list;
		delete[] clip;
	}
	num=numTot=0;
	back=0;
}

void World::addObject(Object *ob) {
	ObjectPtr *tmp=new ObjectPtr[num+1];
	for(int i=0;i<num;i++)
		tmp[i]=obj[i];
	tmp[num]=ob;
	if (num!=0) {
		delete[] obj;
	}
	num++;
	obj=tmp;
	if (numTot!=0) {
		delete[] list;
		delete[] clip;
	}
	numTot=0;
	for(int i=0;i<num;i++)
		numTot+=obj[i]->getNum();
	list=new ElementPtr[numTot];
	clip=new ClipInfo[numTot];
	int cnt=0;
	for(int i=0;i<num;i++)
		for(int j=0;j<obj[i]->getNum();j++)
			list[cnt++]=obj[i]->getElement(j);
}

void World::setBackground(Background *bg) {
	//if (back!=0) delete back;
	if (bg==0) back=0;
	else back=bg;
}

void World::setPerspective(float fov,float zn,float zf) {
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(fov,8.0/6.0,zn,zf);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	gInfo.nPlane=zn;
}


void World::render(Camera *cam) {
	gInfo.cam=cam;
	//Applica sfondo
	if (back!=0) {
		glDisable(GL_DEPTH_TEST);
		glLoadIdentity();
		cam->applyRot();
		back->render();
		glEnable(GL_DEPTH_TEST);
	}
	//Render scena
	glLoadIdentity();
	cam->apply();
	glGetFloatv(GL_MODELVIEW_MATRIX,gInfo.cameraMat);
	setupFrustum();
	//Trasformazioni oggetto
	for(int i=0;i<num;i++)
		obj[i]->transform();
	//Clipping degli element
	for(int i=0;i<numTot;i++)
		clip[i]=list[i]->clipped();
	//Ordinamento degli element
	quickSort(0,numTot-1);
	//Applicazione delle prime 8 luci
	int n=0;
	bool ok=false;
	for(int i=0;i<8;i++)
		glDisable(GL_LIGHT0+i);
	for(int i=numTot-1;(i>=0)&&(n<8);i--)
		if (list[i]->id()==LGHT) {
			if (((LightElement*)list[i])->enabled()) {
					if (!ok) {
						gInfo.lightPos=((LightElement*)list[i])->getPosition();
						ok=true;
					}
				((LightElement*)list[i])->apply(GL_LIGHT0+n);
				n++;
			}
		}
	//Trasformazione luce in obj coords
	for(int i=0;i<num;i++)
		obj[i]->localLight();
	//Disegno degli element
	for(int i=0;i<numTot;i++) {
		if ((!clip[i].clipped)&&(list[i]->getVisible())) list[i]->render();
	}
}

Object* World::getObject(int n) {
	if ((n>=0)&&(n<num))	return obj[n];
	else return 0;
}

int World::getNum() {
	return num;
}

void World::quickSort(int basso,int alto) {
	//dal pi lontano al pi vicino
	ClipInfo tmpI;
	ElementPtr tmpE;
	if (basso<alto) {
		int ba=basso;
		int al=alto+1;
		for(;;) {
			while ((clip[++ba].dist>clip[basso].dist)&&(ba<alto));
			while ((clip[--al].dist<clip[basso].dist)&&(al>basso));
			if (ba<al) {
				tmpI=clip[ba];
				clip[ba]=clip[al];
				clip[al]=tmpI;
				tmpE=list[ba];
				list[ba]=list[al];
				list[al]=tmpE;
			}
			else break;
		}
		tmpI=clip[basso];
		clip[basso]=clip[al];
		clip[al]=tmpI;
		tmpE=list[basso];
		list[basso]=list[al];
		list[al]=tmpE;
		quickSort(basso,al-1);
		quickSort(al+1,alto);
	}
}

void World::setupFrustum() {
	float proj[16];
	float modl[16];
	float clpm[16];
	float t;
	glGetFloatv(GL_PROJECTION_MATRIX,proj);
	glGetFloatv(GL_MODELVIEW_MATRIX,modl);
	clpm[ 0]=modl[ 0]*proj[ 0]+modl[ 1]*proj[ 4]+modl[ 2]*proj[ 8]+modl[ 3]*proj[12];
	clpm[ 1]=modl[ 0]*proj[ 1]+modl[ 1]*proj[ 5]+modl[ 2]*proj[ 9]+modl[ 3]*proj[13];
	clpm[ 2]=modl[ 0]*proj[ 2]+modl[ 1]*proj[ 6]+modl[ 2]*proj[10]+modl[ 3]*proj[14];
	clpm[ 3]=modl[ 0]*proj[ 3]+modl[ 1]*proj[ 7]+modl[ 2]*proj[11]+modl[ 3]*proj[15];
	clpm[ 4]=modl[ 4]*proj[ 0]+modl[ 5]*proj[ 4]+modl[ 6]*proj[ 8]+modl[ 7]*proj[12];
	clpm[ 5]=modl[ 4]*proj[ 1]+modl[ 5]*proj[ 5]+modl[ 6]*proj[ 9]+modl[ 7]*proj[13];
	clpm[ 6]=modl[ 4]*proj[ 2]+modl[ 5]*proj[ 6]+modl[ 6]*proj[10]+modl[ 7]*proj[14];
	clpm[ 7]=modl[ 4]*proj[ 3]+modl[ 5]*proj[ 7]+modl[ 6]*proj[11]+modl[ 7]*proj[15];
	clpm[ 8]=modl[ 8]*proj[ 0]+modl[ 9]*proj[ 4]+modl[10]*proj[ 8]+modl[11]*proj[12];
	clpm[ 9]=modl[ 8]*proj[ 1]+modl[ 9]*proj[ 5]+modl[10]*proj[ 9]+modl[11]*proj[13];
	clpm[10]=modl[ 8]*proj[ 2]+modl[ 9]*proj[ 6]+modl[10]*proj[10]+modl[11]*proj[14];
	clpm[11]=modl[ 8]*proj[ 3]+modl[ 9]*proj[ 7]+modl[10]*proj[11]+modl[11]*proj[15];
	clpm[12]=modl[12]*proj[ 0]+modl[13]*proj[ 4]+modl[14]*proj[ 8]+modl[15]*proj[12];
	clpm[13]=modl[12]*proj[ 1]+modl[13]*proj[ 5]+modl[14]*proj[ 9]+modl[15]*proj[13];
	clpm[14]=modl[12]*proj[ 2]+modl[13]*proj[ 6]+modl[14]*proj[10]+modl[15]*proj[14];
	clpm[15]=modl[12]*proj[ 3]+modl[13]*proj[ 7]+modl[14]*proj[11]+modl[15]*proj[15];
	/* Extract the numbers for the RIGHT plane */
	gInfo.frustum[0][0]=clpm[ 3]-clpm[ 0];
	gInfo.frustum[0][1]=clpm[ 7]-clpm[ 4];
	gInfo.frustum[0][2]=clpm[11]-clpm[ 8];
	gInfo.frustum[0][3]=clpm[15]-clpm[12];
	t=sqrt(gInfo.frustum[0][0]*gInfo.frustum[0][0]+gInfo.frustum[0][1]*gInfo.frustum[0][1]+gInfo.frustum[0][2]*gInfo.frustum[0][2]);
	gInfo.frustum[0][0]/=t;
	gInfo.frustum[0][1]/=t;
	gInfo.frustum[0][2]/=t;
	gInfo.frustum[0][3]/=t;
	/* Extract the numbers for the LEFT plane */
	gInfo.frustum[1][0]=clpm[ 3]+clpm[ 0];
	gInfo.frustum[1][1]=clpm[ 7]+clpm[ 4];
	gInfo.frustum[1][2]=clpm[11]+clpm[ 8];
	gInfo.frustum[1][3]=clpm[15]+clpm[12];
	t=sqrt(gInfo.frustum[1][0]*gInfo.frustum[1][0]+gInfo.frustum[1][1]*gInfo.frustum[1][1]+gInfo.frustum[1][2]*gInfo.frustum[1][2]);
	gInfo.frustum[1][0]/=t;
	gInfo.frustum[1][1]/=t;
	gInfo.frustum[1][2]/=t;
	gInfo.frustum[1][3]/=t;
	/* Extract the BOTTOM plane */
	gInfo.frustum[2][0]=clpm[ 3]+clpm[ 1];
	gInfo.frustum[2][1]=clpm[ 7]+clpm[ 5];
	gInfo.frustum[2][2]=clpm[11]+clpm[ 9];
	gInfo.frustum[2][3]=clpm[15]+clpm[13];
	t=sqrt(gInfo.frustum[2][0]*gInfo.frustum[2][0]+gInfo.frustum[2][1]*gInfo.frustum[2][1]+gInfo.frustum[2][2]*gInfo.frustum[2][2]);
	gInfo.frustum[2][0]/=t;
	gInfo.frustum[2][1]/=t;
	gInfo.frustum[2][2]/=t;
	gInfo.frustum[2][3]/=t;
	/* Extract the TOP plane */
	gInfo.frustum[3][0]=clpm[ 3]-clpm[ 1];
	gInfo.frustum[3][1]=clpm[ 7]-clpm[ 5];
	gInfo.frustum[3][2]=clpm[11]-clpm[ 9];
	gInfo.frustum[3][3]=clpm[15]-clpm[13];
	t=sqrt(gInfo.frustum[3][0]*gInfo.frustum[3][0]+gInfo.frustum[3][1]*gInfo.frustum[3][1]+gInfo.frustum[3][2]*gInfo.frustum[3][2]);
	gInfo.frustum[3][0]/=t;
	gInfo.frustum[3][1]/=t;
	gInfo.frustum[3][2]/=t;
	gInfo.frustum[3][3]/=t;
	/* Extract the FAR plane */
	gInfo.frustum[4][0]=clpm[ 3]-clpm[ 2];
	gInfo.frustum[4][1]=clpm[ 7]-clpm[ 6];
	gInfo.frustum[4][2]=clpm[11]-clpm[10];
	gInfo.frustum[4][3]=clpm[15]-clpm[14];
	t=sqrt(gInfo.frustum[4][0]*gInfo.frustum[4][0]+gInfo.frustum[4][1]*gInfo.frustum[4][1]+gInfo.frustum[4][2]*gInfo.frustum[4][2]);
	gInfo.frustum[4][0]/=t;
	gInfo.frustum[4][1]/=t;
	gInfo.frustum[4][2]/=t;
	gInfo.frustum[4][3]/=t;
	/* Extract the NEAR plane */
	gInfo.frustum[5][0]=clpm[ 3]+clpm[ 2];
	gInfo.frustum[5][1]=clpm[ 7]+clpm[ 6];
	gInfo.frustum[5][2]=clpm[11]+clpm[10];
	gInfo.frustum[5][3]=clpm[15]+clpm[14];
	t=sqrt(gInfo.frustum[5][0]*gInfo.frustum[5][0]+gInfo.frustum[5][1]*gInfo.frustum[5][1]+gInfo.frustum[5][2]*gInfo.frustum[5][2]);
	gInfo.frustum[5][0]/=t;
	gInfo.frustum[5][1]/=t;
	gInfo.frustum[5][2]/=t;
	gInfo.frustum[5][3]/=t;
}


// *****************************
// *****************************
// *** M e s h E l e m e n t ***
// *****************************
// *****************************

MeshElement::MeshElement() {
	list=0;
	ready=false;
}

MeshElement::~MeshElement() {
	kill();
}

void MeshElement::kill() {
	if (!ready) return;
	delete[] vert;
	delete[] mapc;
	delete[] face;
	delete[] norm;
	glDeleteLists(list,1);
	mat.kill();
	ready=false;
}

void MeshElement::create(int nnv,int nnf,Vector *vv,MapCoord *mm,Triangle *tt) {
	if (ready) return;
	nv=nnv;nf=nnf;
	vert=vv;mapc=mm;face=tt;
	if (mapc==0) mapc=new MapCoord[nv];
	int i,j;
	// Rotate X 90
	for(i=0;i<nv;i++) {
		float tmp=vert[i].y;
		vert[i].y=vert[i].z;
		vert[i].z=-tmp;
	}
	// Sphere for clipping
	center=Vector(0.0,0.0,0.0);
	for(i=0;i<nv;i++)
		center.add(vert[i]);
	center=mul(1/nv,center);
	for(i=0;i<nv;i++)
		if (ray<length(sub(vert[i],center))) ray=length(sub(vert[i],center));
	// Generate normals for smooth shading
	norm=new Vector[nv];
	Vector cv;
	for(i=0;i<nv;i++) {
		norm[i].x=norm[i].y=norm[i].z=0;
		// Search for overlapping vertices
		cv.x=vert[i].x;
		cv.y=vert[i].y;
		cv.z=vert[i].z;
		for(j=0;j<nv;j++) {
			if ((cv.x==vert[j].x)&&(cv.y==vert[j].y)&&(cv.z==vert[j].z)) {
				for(int f=0;f<nf;f++)
					if ((face[f].a==j)||(face[f].b==j)||(face[f].c==j)) {
						norm[i]=add(norm[i],cross(sub(vert[face[f].b],vert[face[f].a]),sub(vert[face[f].c],vert[face[f].a])));
					}
			}
		}
		norm[i].normalize();
	}
	// Generate display list
	list=glGenLists(1);
	glNewList(list,GL_COMPILE);
	int a,b,c;
	for(i=0;i<nf;i++) {
		glBegin(GL_TRIANGLES);
		a=face[i].a;b=face[i].b;c=face[i].c;
		glNormal3f(norm[a].x,norm[a].y,norm[a].z);
		glTexCoord2f(mapc[a].u,mapc[a].v);
		glVertex3f(vert[a].x,vert[a].y,vert[a].z);
		glNormal3f(norm[b].x,norm[b].y,norm[b].z);
		glTexCoord2f(mapc[b].u,mapc[b].v);
		glVertex3f(vert[b].x,vert[b].y,vert[b].z);
		glNormal3f(norm[c].x,norm[c].y,norm[c].z);
		glTexCoord2f(mapc[c].u,mapc[c].v);
		glVertex3f(vert[c].x,vert[c].y,vert[c].z);
		glEnd();
	}
	glEndList();
	ready=true;
}

void MeshElement::setMaterial(Material m) {
	mat=m;
}

void MeshElement::setTransparency(bool tp,float al) {
	if (tp) {
		mat.opaque=false;
		mat.setTransparency(al);
	}
	else {
		mat.opaque=true;
		mat.setTransparency(1.0f);
	}
}

void MeshElement::setDiffuseMap(char *fname) {
	// Linear filtering, mipmapping
	mat.setDiffuseMap(fname,DETAIL2);
}

void MeshElement::render() {
	if (!ready) return;
	glPushMatrix();
	glMultMatrixf(info->objectMat);
	glTranslatef(posx,posy,posz);
	// Attiva Antialiasing
	if ((mat.opaque)&&(mat.bump<0.0)&&(antialias)) glEnable(GL_POLYGON_SMOOTH);
	// Attiva Trasparenza
	if (!mat.opaque) {
		glDisable(GL_CULL_FACE);
		glDepthMask(GL_FALSE);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE);
	}
	// Attiva modalit
	if (mode==FLAT) glShadeModel(GL_FLAT);
	else if (mode==WIREFRAME) {
		glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
		glDisable(GL_LIGHTING);
	}
	mat.applyColor();
	mat.applyDiffuse();
	glCallList(list);
	// Disattiva modalit
	if (mode==FLAT) glShadeModel(GL_SMOOTH);
	else if (mode==WIREFRAME) {
		glEnable(GL_LIGHTING);
		glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
	}
	// Disattiva Trasparenza
	if (!mat.opaque) {
		glEnable(GL_CULL_FACE);
		glDepthMask(GL_TRUE);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	}
	// Disattiva Antialiasing
	if ((mat.opaque)&&(mat.bump<0.0)&&(antialias)) glDisable(GL_POLYGON_SMOOTH);
	glPopMatrix();
}

void MeshElement::renderFlat() {
	if (!ready) return;
	glPushMatrix();
	glMultMatrixf(info->objectMat);
	glTranslatef(posx,posy,posz);
	glCallList(list);
	glPopMatrix();
}

void MeshElement::setPosition(float px,float py,float pz) {
	posx=px;
	posy=py;
	posz=pz;
}

ClipInfo MeshElement::clipped() {
	ClipInfo ci;
	ci.clipped=false;
	Vector p(posx,posy,posz);
	Vector objx(info->objectMat[0],info->objectMat[1],info->objectMat[2]);
	Vector objy(info->objectMat[4],info->objectMat[5],info->objectMat[6]);
	Vector objz(info->objectMat[8],info->objectMat[9],info->objectMat[10]);
	Matrix m(objx,objy,objz);
	p=m.xForm(p);
	float d;
	for(int i=0;i<6;i++) {
		d=gInfo.frustum[i][0]*(p.x+center.x+info->objectPos.x)+
		  gInfo.frustum[i][1]*(p.y+center.y+info->objectPos.y)+
		  gInfo.frustum[i][2]*(p.z+center.z+info->objectPos.z)+
		  gInfo.frustum[i][3];
		if (d<=-ray) ci.clipped=true;
	}
	ci.dist=d+ray;
	return ci;
}

int MeshElement::id() {
	return MESH;
}


// *******************************
// *** L i g h t E l e m e n t ***
// *******************************

LightElement::LightElement() {
	posx=posy=posz=0.0;
	ldir.set(0.0,-1.0,0.0);
	w0=1.0;
	cutoff=180;
	expon=1.0;
	cf=1.0;lf=qf=0.0;
	lcol.set(1.0,1.0,1.0);
	nFlare=0;
	vis=en1=en2=en3=false;
	ready2=ready3=false;
}

LightElement::~LightElement() {
	kill();
}

void LightElement::create(char *fname,float f) {
	if (ready2) return;
	if (fname==0) return;
	dim=f;
	ray=dim;
	map.setDiffuseMap(fname,DETAIL1);
	ready2=true;
}

void LightElement::kill() {
	if (!(ready2||ready3)) return;
	if (ready2) {
		map.kill();
		en1=en2=false;
		ready2=false;
	}
	if (ready3) {
		for(int i=0;i<nFlare;i++)
			flare[i].map.kill();
		delete[] flare;
		nFlare=0;
		ready3=en3=false;
	}
}

void LightElement::apply(int nl) {
	if (!en1) {
		glDisable(nl);
		return;
	}
	glEnable(nl);
	glPushMatrix();
	glMultMatrixf(info->objectMat);
	glTranslatef(posx,posy,posz);
	GLfloat pos[]={0.0,0.0,0.0,w0};
	GLfloat dir[]={ldir.x,ldir.y,ldir.z};
	glLighti(nl,GL_SPOT_CUTOFF,cutoff);
	glLightf(nl,GL_SPOT_EXPONENT,expon);
	glLighti(nl,GL_CONSTANT_ATTENUATION,cf);
	glLighti(nl,GL_LINEAR_ATTENUATION,lf);
	glLighti(nl,GL_QUADRATIC_ATTENUATION,qf);
	GLfloat amb[]={0.0,0.0,0.0,1.0};
	GLfloat col[]={lcol.x,lcol.y,lcol.z,1.0};
	glLightfv(nl,GL_AMBIENT,amb);
	glLightfv(nl,GL_DIFFUSE,col);
	glLightfv(nl,GL_SPECULAR,col);
	glLightfv(nl,GL_POSITION,pos);
	glLightfv(nl,GL_SPOT_DIRECTION,dir);
	glPopMatrix();
}

void LightElement::render() {
	float cam[16];
	glGetFloatv(GL_MODELVIEW_MATRIX,cam);
	glPushMatrix();
	glMultMatrixf(info->objectMat);
	glTranslatef(posx,posy,posz);
	float cm[16];
	glGetFloatv(GL_MODELVIEW_MATRIX,cm);
	Vector p1(cm[0]*dim/2,cm[4]*dim/2,cm[8]*dim/2);
	Vector p2(cm[1]*dim/2,cm[5]*dim/2,cm[9]*dim/2);
	glPushAttrib(GL_FOG_BIT);
	glDisable(GL_FOG);
	if (ready2&&en2) {
		glDepthMask(GL_FALSE);
		glDisable(GL_LIGHTING);
		glEnable(GL_COLOR_MATERIAL);
		glBlendFunc(GL_ONE,GL_ONE);
		glColor4f(lcol.x,lcol.y,lcol.z,0.0);
		map.apply();
		glBegin(GL_QUADS);
			glTexCoord2f(0.0,0.0);
			glVertex3f(-p1.x-p2.x,-p1.y-p2.y,-p1.z-p2.z);
			glTexCoord2f(1.0,0.0);
			glVertex3f(+p1.x-p2.x,+p1.y-p2.y,+p1.z-p2.z);
			glTexCoord2f(1.0,1.0);
			glVertex3f(+p1.x+p2.x,+p1.y+p2.y,+p1.z+p2.z);
			glTexCoord2f(0.0,1.0);
			glVertex3f(-p1.x+p2.x,-p1.y+p2.y,-p1.z+p2.z);
		glEnd();
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		glDisable(GL_COLOR_MATERIAL);
		glEnable(GL_LIGHTING);
		glDepthMask(GL_TRUE);
	}
	if (ready3&&en3) {
		Vector lp(posx,posy,posz);
		Vector objx(info->objectMat[0],info->objectMat[1],info->objectMat[2]);
		Vector objy(info->objectMat[4],info->objectMat[5],info->objectMat[6]);
		Vector objz(info->objectMat[8],info->objectMat[9],info->objectMat[10]);
		Matrix m(objx,objy,objz);
		lp=m.xForm(lp);
		lp=add(lp,info->objectPos);
			
		Vector vd=gInfo.cam->getViewDir();
		Vector ep=gInfo.cam->getEyePoint();

		vd.normalize();
		Vector cnear=add(ep,mul(2*gInfo.nPlane,vd));
		Vector lighteye=sub(lp,ep); //a che serve?? Luminosit?
		float dotp=pow(dot(normalize(sub(lp,cnear)),vd),15.0f);
		Vector axis=sub(cnear,lp);
		float len=length(axis);
		Vector pos;
		glDisable(GL_LIGHTING);
		glDisable(GL_DEPTH_TEST);
		glEnable(GL_COLOR_MATERIAL);
		glBlendFunc(GL_ONE,GL_ONE);
		glPushMatrix();
		glLoadIdentity();
		glLoadMatrixf(cam);
		Vector fp;
		// una matrice diversa, ricalcolo p1 e p2
		p1.set(cam[0]*dim/2,cam[4]*dim/2,cam[8]*dim/2);
		p2.set(cam[1]*dim/2,cam[5]*dim/2,cam[9]*dim/2);
		Vector p1old=p1;
		Vector p2old=p2;
		float lum=dotp/(0.02f*len+1.0f);
		for(int i=0;i<nFlare;i++) {
			glColor3f(flare[i].rf*lum*lcol.x,flare[i].gf*lum*lcol.y,flare[i].bf*lum*lcol.z);
			flare[i].map.apply();
			fp=mul(flare[i].pos,axis);
			fp=add(lp,fp);
			Vector d=sub(fp,ep);
			if ((i!=0)&&(i!=1)) {
			   p1=mul(length(d)*flare[i].dim/(len+gInfo.nPlane),p1old);
			   p2=mul(length(d)*flare[i].dim/(len+gInfo.nPlane),p2old);
		   }
			glBegin(GL_QUADS);
				glTexCoord2f(0.0,0.0);
				glVertex3f(fp.x-p1.x-p2.x,fp.y-p1.y-p2.y,fp.z-p1.z-p2.z);
				glTexCoord2f(1.0,0.0);
				glVertex3f(fp.x+p1.x-p2.x,fp.y+p1.y-p2.y,fp.z+p1.z-p2.z);
				glTexCoord2f(1.0,1.0);
				glVertex3f(fp.x+p1.x+p2.x,fp.y+p1.y+p2.y,fp.z+p1.z+p2.z);
				glTexCoord2f(0.0,1.0);
				glVertex3f(fp.x-p1.x+p2.x,fp.y-p1.y+p2.y,fp.z-p1.z+p2.z);
			glEnd();
		}
		glPopMatrix();
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_DEPTH_TEST);
		glDisable(GL_COLOR_MATERIAL);
		glEnable(GL_LIGHTING);
	}
	glPopAttrib();
	if (vis) {
		glDisable(GL_TEXTURE_2D);
	   glDisable(GL_LIGHTING);
		glDisable(GL_BLEND);
		glEnable(GL_COLOR_MATERIAL);
		glPointSize(5.0);
		glLineWidth(2.0);
		glColor3f(lcol.x,lcol.y,lcol.z);
		glBegin(GL_POINTS);
			glVertex3f(0,0,0);
		glEnd();
		glBegin(GL_LINE_STRIP);
			glVertex3f(0,0,0);
			glVertex3f(0+ldir.x,0+ldir.y,0+ldir.z);
		glEnd();
		glDisable(GL_COLOR_MATERIAL);
		glEnable(GL_BLEND);
	   glEnable(GL_LIGHTING);
	}
	glPopMatrix();
}

void LightElement::renderFlat() {
	return;
}

void LightElement::setPosition(float px,float py,float pz) {
	posx=px;
	posy=py;
	posz=pz;
}

Vector LightElement::getPosition() {
	float *om=info->objectMat;
	Vector tmp;
	tmp.x=om[0]*posx+om[4]*posy+om[8]*posz+om[12];
	tmp.y=om[1]*posx+om[5]*posy+om[9]*posz+om[13];
	tmp.z=om[2]*posx+om[6]*posy+om[10]*posz+om[14];
	return tmp;
}

void LightElement::setType(int a,float e,Vector &d,int tp) {
	if (a<0) a=0;
	if ((a>90)&&(a!=180)) a=90;
	cutoff=a;
	expon=e;
	ldir=d;
	if (tp==DIRECTIONAL) w0=0.0;
	else w0=1.0;
}

void LightElement::setAttenuation(float c,float l,float q) {
	if (c<0.0) c=1.0;
	if (l<0.0) l=0.0;
	if (q<0.0) q=0.0;
	cf=c;lf=l;qf=q;
}

void LightElement::setColor(float r,float g,float b) {
	lcol.x=r;
	lcol.y=g;
	lcol.z=b;
}

void LightElement::setFlare(char *fname,int nf) {
	if (ready3) return;
	if (nf<3) return;
	if (nFlare!=0) delete[] flare;
	nFlare=nf;
	char fn[50];
	strcpy(fn,fname);
	flare=new Flare[nFlare];
	int l=0;
	while(fn[l++]!=0);
	for(int i=0;i<nFlare;i++) {
		fn[l-6]=(char)('0'+i);
		flare[i].map.setDiffuseMap(fn,DETAIL1);
		//flare[i].pos=1.0f-1.0f/((i-1)*(20.0f/(nFlare-2))+0.01f);
		flare[i].pos=pow(float(i)/float(nFlare-1),0.01f)+(float(rand()%10)/1000.0f);
		flare[i].dim=float(rand()%100)/200.0f+0.1f;
		flare[i].rf=1.0f-float(rand()%100)/200.0f;
		flare[i].gf=1.0f-float(rand()%100)/200.0f;
		flare[i].bf=1.0f-float(rand()%100)/200.0f;
	}
	//flare i=0,1 >> shine
	flare[0].pos=flare[1].pos=0.0f;
	flare[0].dim=flare[1].dim=dim/6.0f;
	//last flare
	flare[nFlare-1].pos=1.0f+float(rand()%10)/1000.0f;
	ready3=true;
}

bool LightElement::enabled() {
	return en1;
}

void LightElement::enable(bool b1,bool b2,bool b3) {
	en1=b1;
	en2=b2;
	en3=b3;
}

void LightElement::visible(bool b) {
	vis=b;
}

ClipInfo LightElement::clipped() {
	ClipInfo ci;
	ci.clipped=false;
	Vector p(posx,posy,posz);
	Vector objx(info->objectMat[0],info->objectMat[1],info->objectMat[2]);
	Vector objy(info->objectMat[4],info->objectMat[5],info->objectMat[6]);
	Vector objz(info->objectMat[8],info->objectMat[9],info->objectMat[10]);
	Matrix m(objx,objy,objz);
	p=m.xForm(p);
	float d;
	for(int i=0;i<6;i++) {
		d=gInfo.frustum[i][0]*(p.x+center.x+info->objectPos.x)+
		  gInfo.frustum[i][1]*(p.y+center.y+info->objectPos.y)+
		  gInfo.frustum[i][2]*(p.z+center.z+info->objectPos.z)+
		  gInfo.frustum[i][3];
		if (d<=-ray) ci.clipped=true;
	}
	ci.dist=fabs(d)/1000.0f;
	return ci;
}

int LightElement::id() {
	return LGHT;
}


// *******************************
// *** F l u i d E l e m e n t ***
// *******************************

FluidElement::FluidElement() {
	ready=false;
	iso=6.0;
	mass=0.2;
	dist=5;
	visc=0.95;
	xmin=ymin=zmin=-1000;
	xmax=ymax=zmax=1000;	
	gravity.set(0.0,-200,0.0);
	drawAtm=drawGrd=false;
}

FluidElement::~FluidElement() {
	kill();
}
	
void FluidElement::create(int na,int gs,float gx,float gy,float gz) {
	if (na<=0) return;
	if (ready) kill();
	num=na;
	gSize=gs;
	atoms=new Vector[num];
	speed=new Vector[num];
	field=new Vector[num];
	gridp=new GridPoint[gSize*gSize*gSize];
	cells=new GridCell[(gSize-1)*(gSize-1)*(gSize-1)];
	vList=new Vector[12];
	nList=new Vector[12];
	center.set(0.0,0.0,0.0);
	ray=sqrt(gx*gx+gy*gy+gz*gz);
	int idx;
	for(int cx=0;cx<gSize;cx++)
		for(int cy=0;cy<gSize;cy++)
			for(int cz=0;cz<gSize;cz++) {
				idx=cx+gSize*(cy+gSize*cz);
				gridp[idx].pos.x=gx*(2.0f*float(cx)/gSize-1.0f);
				gridp[idx].pos.y=gy*(2.0f*float(cy)/gSize-1.0f);
				gridp[idx].pos.z=gz*(1.0f-2.0f*float(cz)/gSize);
			}
	for(int cx=0;cx<gSize-1;cx++)
		for(int cy=0;cy<gSize-1;cy++)
			for(int cz=0;cz<gSize-1;cz++) {
				idx=cx+(gSize-1)*(cy+(gSize-1)*cz);
				cells[idx].gPoint[0]=&gridp[(cx)+gSize*((cy)+gSize*(cz))];
				cells[idx].gPoint[1]=&gridp[(cx+1)+gSize*((cy)+gSize*(cz))];
				cells[idx].gPoint[2]=&gridp[(cx+1)+gSize*((cy)+gSize*(cz+1))];
				cells[idx].gPoint[3]=&gridp[(cx)+gSize*((cy)+gSize*(cz+1))];
				cells[idx].gPoint[4]=&gridp[(cx)+gSize*((cy+1)+gSize*(cz))];
				cells[idx].gPoint[5]=&gridp[(cx+1)+gSize*((cy+1)+gSize*(cz))];
				cells[idx].gPoint[6]=&gridp[(cx+1)+gSize*((cy+1)+gSize*(cz+1))];
				cells[idx].gPoint[7]=&gridp[(cx)+gSize*((cy+1)+gSize*(cz+1))];
			}	
	for(int i=0;i<num;i++) {
		atoms[i].set(0.5-float(rand())/RAND_MAX,0.5-float(rand())/RAND_MAX,0.5-float(rand())/RAND_MAX);
	}
	ready=true;
}

void FluidElement::kill() {
	if (ready) {
		delete[] atoms;
		delete[] speed;
		delete[] field;
		delete[] gridp;
		delete[] cells;
		delete[] vList;
		delete[] nList;
		num=0;
		gSize=0;
		mat.kill();
	}
	ready=false;
}

void FluidElement::setEnvironmentMap(char *fname) {
	// Linear filtering, no mipmapping
	mat.setDiffuseMap(fname,DETAIL1);
}

void FluidElement::render() {
	if (!ready) return;
	glPushMatrix();
	glMultMatrixf(info->objectMat);
	glTranslatef(posx,posy,posz);
	
	if (antialias) glEnable(GL_POLYGON_SMOOTH);
	if (mode==FLAT) glShadeModel(GL_FLAT);
	else if (mode==WIREFRAME) glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);

	glEnable(GL_TEXTURE_GEN_S);
	glEnable(GL_TEXTURE_GEN_T);
	glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);

	glEnable(GL_NORMALIZE);
	
	mat.applyColor();
	mat.applyDiffuse();
	glBegin(GL_TRIANGLES);
	int dim=(gSize-1)*(gSize-1)*(gSize-1);
	for(int i=0;i<dim;i++)
		surface(cells[i]);
	glEnd();
	
	glDisable(GL_NORMALIZE);
	
	glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
	glDisable(GL_TEXTURE_GEN_S);
	glDisable(GL_TEXTURE_GEN_T);			

	if (drawAtm) {
		glEnable(GL_COLOR_MATERIAL);
		glEnable(GL_POINT_SMOOTH);
		glDisable(GL_LIGHTING);
		glDisable(GL_TEXTURE_2D);
		glPointSize(5.0f);
		glColor3f(0.3,0.3,1.0);
		glBegin(GL_POINTS);
		for(int i=0;i<num;i++)
				glVertex3f(atoms[i].x,atoms[i].y,atoms[i].z);
		glEnd();
		glEnable(GL_TEXTURE_2D);
		glEnable(GL_LIGHTING);
		glDisable(GL_POINT_SMOOTH);
		glDisable(GL_COLOR_MATERIAL);
	}

	if (drawGrd) {
		glEnable(GL_COLOR_MATERIAL);
		glEnable(GL_POINT_SMOOTH);
		glDisable(GL_LIGHTING);
		glDisable(GL_TEXTURE_2D);
		GridPoint *gp;
		glPointSize(1.0f);
		glColor3f(0.3,0.3,1.0);
		glBegin(GL_POINTS);
		for(int cx=0;cx<gSize;cx++)
			for(int cy=0;cy<gSize;cy++)
				for(int cz=0;cz<gSize;cz++) {
					gp=gridp+cx+gSize*(cy+gSize*cz);
					glVertex3f(gp->pos.x,gp->pos.y,gp->pos.z);
				}
		glEnd();
		glEnable(GL_TEXTURE_2D);
		glEnable(GL_LIGHTING);
		glDisable(GL_POINT_SMOOTH);
		glDisable(GL_COLOR_MATERIAL);
	}
	
	if (mode==FLAT) glShadeModel(GL_SMOOTH);
	else if (mode==WIREFRAME) glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
	if (antialias) glDisable(GL_POLYGON_SMOOTH);

	glPopMatrix();
}

void FluidElement::renderFlat() {

}

void FluidElement::updateAtoms(float tm) {
	if (!ready) return;
	Vector seg;
	float dst;
	float fld;
	tm=tm;
	for(int cr=0;cr<num;cr++) {
		field[cr].set(0.0,0.0,0.0);
		for(int i=0;i<num;i++) {
			if (cr==i) continue;
			seg=sub(atoms[cr],atoms[i]);
			dst=length(seg);
			if (dst<0.05) dst=0.05;
			dst=1.0/dst;
			// funzione potenziale
			//fld=dsti*(dsti-1); // 1/x^2 - 1/x
			fld=2*mass*(dst*dst*(dst-1/dist));
			//fld=mass*10*(dst*dst*(dst-2.0));
			field[cr].add(mul(fld,seg));
		}
		field[cr].add(mul(mass,gravity));
	}
	float tmm=tm/mass;
	//float coeff=0.93;
	//float coeff=0.95;
	float loss=0.8*visc;
	for(int cr=0;cr<num;cr++) {
		speed[cr]=add(mul(visc,speed[cr]),mul(tmm,field[cr])); // v=vi+F*t/m
		atoms[cr]=add(atoms[cr],mul(tm,speed[cr])); // p=pi+v*t
		if (atoms[cr].x<xmin) {
			atoms[cr].x=2*xmin-atoms[cr].x;
			speed[cr].x=loss*fabs(speed[cr].x);
		}
		if (atoms[cr].x>xmax) {
			atoms[cr].x=2*xmax-atoms[cr].x;
			speed[cr].x=-loss*fabs(speed[cr].x);
		}
		if (atoms[cr].y<ymin) {
			atoms[cr].y=2*ymin-atoms[cr].y;
			speed[cr].y=loss*fabs(speed[cr].y);
		}
		if (atoms[cr].y>ymax) {
			atoms[cr].y=2*ymax-atoms[cr].y;
			speed[cr].y=-loss*fabs(speed[cr].y);
		}
		if (atoms[cr].z<zmin) {
			atoms[cr].z=2*zmin-atoms[cr].z;
			speed[cr].z=loss*fabs(speed[cr].z);
		}
		if (atoms[cr].z>zmax) {
			atoms[cr].z=2*zmax-atoms[cr].z;
			speed[cr].z=-loss*fabs(speed[cr].z);
		}
	}
	
	// Update field on grid points
	float dx,dy,dz;
	GridPoint *gp;
	int tmp=gSize*gSize*gSize;
	for(int i=0;i<tmp;i++) {
		gp=gridp+i;
		gp->val=0;
		for(int cr=0;cr<num;cr++) {
			dx=gp->pos.x-atoms[cr].x;
			dy=gp->pos.y-atoms[cr].y;
			dz=gp->pos.z-atoms[cr].z;
			gp->val+=1.0/(dx*dx+dy*dy+dz*dz);
		}
	}
	for(int x=1;x<gSize-1;x++)
		for(int y=1;y<gSize-1;y++)
			for(int z=1;z<gSize-1;z++) {
				tmp=x+gSize*(y+gSize*z);
				gridp[tmp].norm.x=gridp[tmp-1].val-gridp[tmp+1].val;
				gridp[tmp].norm.y=gridp[tmp-gSize].val-gridp[tmp+gSize].val;
				gridp[tmp].norm.z=gridp[tmp+gSize*gSize].val-gridp[tmp-gSize*gSize].val;
			}
}

void FluidElement::visAtoms(bool va) {
	drawAtm=va;
}

void FluidElement::visGrid(bool vg) {
	drawGrd=vg;
}

void FluidElement::setAmbColor(float rr,float gg,float bb,float aa) {
	mat.setAmbient(rr,gg,bb,aa);
}

void FluidElement::setDifColor(float rr,float gg,float bb,float aa) {
	mat.setDiffuse(rr,gg,bb,aa);
}

void FluidElement::setSpcColor(float rr,float gg,float bb,float aa) {
	mat.setSpecular(rr,gg,bb,aa);
}

void FluidElement::setEmsColor(float rr,float gg,float bb,float aa) {
	mat.setEmissive(rr,gg,bb,aa);
}

void FluidElement::setLevel(double l) {
	iso=l;
}

void FluidElement::setGravity(Vector gr) {
	gravity=gr;
}

void FluidElement::setMass(float m) {
	mass=m;
}

void FluidElement::setDistance(float d) {
	dist=d;
}

void FluidElement::setViscosity(float v) {
	visc=v;
}

void FluidElement::setLimits(float x1,float x2,float y1,float y2,float z1,float z2) {
	xmin=x1;xmax=x2;
	ymin=y1;ymax=y2;
	zmin=z1;zmax=z2;	
}

void FluidElement::setPosition(float px,float py,float pz) {
	posx=px;
	posy=py;
	posz=pz;
}

void FluidElement::getAtoms(Vector* &at,Vector* &sp,int &na) {
	at=atoms;
	sp=speed;
	na=num;
}

ClipInfo FluidElement::clipped() {
	ClipInfo ci;
	ci.clipped=false;
	Vector p(posx,posy,posz);
	Vector objx(info->objectMat[0],info->objectMat[1],info->objectMat[2]);
	Vector objy(info->objectMat[4],info->objectMat[5],info->objectMat[6]);
	Vector objz(info->objectMat[8],info->objectMat[9],info->objectMat[10]);
	Matrix m(objx,objy,objz);
	p=m.xForm(p);
	float d;
	for(int i=0;i<6;i++) {
		d=gInfo.frustum[i][0]*(p.x+center.x+info->objectPos.x)+
		  gInfo.frustum[i][1]*(p.y+center.y+info->objectPos.y)+
		  gInfo.frustum[i][2]*(p.z+center.z+info->objectPos.z)+
		  gInfo.frustum[i][3];
		if (d<=-ray) ci.clipped=true;
	}
	ci.dist=gInfo.nPlane;
	return ci;	
}

int FluidElement::id() {
	return FLUI;
}

int FluidElement::surface(GridCell &grid) {
	int nTri;
	int cubeInd;
   // Determine the index into the edge table which
   // tells us which vertices are inside of the surface
   cubeInd=0;
   if (grid.gPoint[0]->val<iso) cubeInd|=1;
   if (grid.gPoint[1]->val<iso) cubeInd|=2;
   if (grid.gPoint[2]->val<iso) cubeInd|=4;
   if (grid.gPoint[3]->val<iso) cubeInd|=8;
   if (grid.gPoint[4]->val<iso) cubeInd|=16;
   if (grid.gPoint[5]->val<iso) cubeInd|=32;
   if (grid.gPoint[6]->val<iso) cubeInd|=64;
   if (grid.gPoint[7]->val<iso) cubeInd|=128;

   // Cube is entirely in/out of the surface
   if (edgeTable[cubeInd]==0) return 0;
      
   // Find the vertices where the surface intersects the cube
   if (edgeTable[cubeInd]&   1) VertexInterp(grid.gPoint[0],grid.gPoint[1],vList[0],nList[0]);
   if (edgeTable[cubeInd]&   2) VertexInterp(grid.gPoint[1],grid.gPoint[2],vList[1],nList[1]);
   if (edgeTable[cubeInd]&   4) VertexInterp(grid.gPoint[2],grid.gPoint[3],vList[2],nList[2]);
   if (edgeTable[cubeInd]&   8) VertexInterp(grid.gPoint[3],grid.gPoint[0],vList[3],nList[3]);
   if (edgeTable[cubeInd]&  16) VertexInterp(grid.gPoint[4],grid.gPoint[5],vList[4],nList[4]);
   if (edgeTable[cubeInd]&  32) VertexInterp(grid.gPoint[5],grid.gPoint[6],vList[5],nList[5]);
   if (edgeTable[cubeInd]&  64) VertexInterp(grid.gPoint[6],grid.gPoint[7],vList[6],nList[6]);
   if (edgeTable[cubeInd]& 128) VertexInterp(grid.gPoint[7],grid.gPoint[4],vList[7],nList[7]);
   if (edgeTable[cubeInd]& 256) VertexInterp(grid.gPoint[0],grid.gPoint[4],vList[8],nList[8]);
   if (edgeTable[cubeInd]& 512) VertexInterp(grid.gPoint[1],grid.gPoint[5],vList[9],nList[9]);
   if (edgeTable[cubeInd]&1024) VertexInterp(grid.gPoint[2],grid.gPoint[6],vList[10],nList[10]);
   if (edgeTable[cubeInd]&2048) VertexInterp(grid.gPoint[3],grid.gPoint[7],vList[11],nList[11]);

   // Create the triangle
   nTri=0;
   Vector *v1,*v2,*v3;
   Vector *n1,*n2,*n3;
   for(int i=0;triTable[cubeInd][i]!=-1;i+=3) {
      v1=vList+triTable[cubeInd][i+0];
      n1=nList+triTable[cubeInd][i+0];
      v2=vList+triTable[cubeInd][i+1];
      n2=nList+triTable[cubeInd][i+1];
      v3=vList+triTable[cubeInd][i+2];
      n3=nList+triTable[cubeInd][i+2];
      glNormal3f(n1->x,n1->y,n1->z);
      glVertex3f(v1->x,v1->y,v1->z);
      glNormal3f(n2->x,n2->y,n2->z);
      glVertex3f(v2->x,v2->y,v2->z);
      glNormal3f(n3->x,n3->y,n3->z);
      glVertex3f(v3->x,v3->y,v3->z);
      nTri++;
   }
   return nTri;
}

void FluidElement::VertexInterp(GridPoint *p1,GridPoint *p2,Vector &v,Vector &n) {
	double mu;
	if (fabs(iso-p1->val)<0.00001) {
		v=p1->pos;
		return;
	}
	if (fabs(iso-p2->val)<0.00001) {
		v=p2->pos;
		return;
	}
	if (fabs(p1->val-p2->val)<0.00001) {
		v=p1->pos;
		return;
	}
	mu=(iso-p1->val)/(p2->val-p1->val);
	v.x=p1->pos.x+mu*(p2->pos.x-p1->pos.x);
	v.y=p1->pos.y+mu*(p2->pos.y-p1->pos.y);
	v.z=p1->pos.z+mu*(p2->pos.z-p1->pos.z);
	n.x=p1->norm.x+mu*(p2->norm.x-p1->norm.x);
	n.y=p1->norm.y+mu*(p2->norm.y-p1->norm.y);
	n.z=p1->norm.z+mu*(p2->norm.z-p1->norm.z);
}



// *******************************************
// *******************************************
// ****    2  D    O  B  J  E  C  T  S    ****
// *******************************************
// *******************************************


// *****************************
// *****************************
// ***   T e x Q u a d 2 D   ***
// *****************************
// *****************************

TexQuad2D::TexQuad2D() {
	ready=false;
	bright=1.0;
	posX=200;posY=150;
	sizeX=400;sizeY=300;
	texC[0].u=0;texC[0].v=0;
	texC[1].u=1;texC[1].v=0;
	texC[2].u=1;texC[2].v=1;
	texC[3].u=0;texC[3].v=1;
	vis=true;
}

TexQuad2D::~TexQuad2D() {
	kill();
}

void TexQuad2D::create(char *fname) {
	if (ready) kill();
	tex.setDiffuseMap(fname,DETAIL1);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
	ready=true;
}

void TexQuad2D::kill() {
	if (!ready) return;
	tex.kill();
	ready=false;
}

void TexQuad2D::render() {
	if (!vis) return;
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE,GL_ONE);
	glEnable(GL_TEXTURE_2D);
	//glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_REPLACE);

	tex.apply();
	glColor3f(bright,bright,bright);
	glBegin(GL_QUADS);
		glTexCoord2f(texC[0].u,texC[0].v);
		glVertex2f(posX,posY);
		glTexCoord2f(texC[1].u,texC[1].v);
		glVertex2f(posX+sizeX,posY);
		glTexCoord2f(texC[2].u,texC[2].v);
		glVertex2f(posX+sizeX,posY+sizeY);
		glTexCoord2f(texC[3].u,texC[3].v);
		glVertex2f(posX,posY+sizeY);		
	glEnd();
	
	//glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
	glDisable(GL_TEXTURE_2D);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_BLEND);
}

void TexQuad2D::setSize(float sx,float sy) {
	sizeX=sx;sizeY=sy;
}

void TexQuad2D::setPosition(float px,float py) {
	posX=px;posY=py;
}

void TexQuad2D::setTexCoord(int nv,float tu,float tv) {
	if ((nv<0)||(nv>3)) return;
	texC[nv].u=tu;
	texC[nv].v=tv;
}

void TexQuad2D::setBrightness(float b) {
	if (b>1.0) b=1.0;
	if (b<0.0) b=0.0;
	bright=b;
}

void TexQuad2D::setVisible(bool b) {
	vis=b;
}

// Lookup tables per FluidElement
const int edgeTable[256]={
0x0  , 0x109, 0x203, 0x30a, 0x406, 0x50f, 0x605, 0x70c,
0x80c, 0x905, 0xa0f, 0xb06, 0xc0a, 0xd03, 0xe09, 0xf00,
0x190, 0x99 , 0x393, 0x29a, 0x596, 0x49f, 0x795, 0x69c,
0x99c, 0x895, 0xb9f, 0xa96, 0xd9a, 0xc93, 0xf99, 0xe90,
0x230, 0x339, 0x33 , 0x13a, 0x636, 0x73f, 0x435, 0x53c,
0xa3c, 0xb35, 0x83f, 0x936, 0xe3a, 0xf33, 0xc39, 0xd30,
0x3a0, 0x2a9, 0x1a3, 0xaa , 0x7a6, 0x6af, 0x5a5, 0x4ac,
0xbac, 0xaa5, 0x9af, 0x8a6, 0xfaa, 0xea3, 0xda9, 0xca0,
0x460, 0x569, 0x663, 0x76a, 0x66 , 0x16f, 0x265, 0x36c,
0xc6c, 0xd65, 0xe6f, 0xf66, 0x86a, 0x963, 0xa69, 0xb60,
0x5f0, 0x4f9, 0x7f3, 0x6fa, 0x1f6, 0xff , 0x3f5, 0x2fc,
0xdfc, 0xcf5, 0xfff, 0xef6, 0x9fa, 0x8f3, 0xbf9, 0xaf0,
0x650, 0x759, 0x453, 0x55a, 0x256, 0x35f, 0x55 , 0x15c,
0xe5c, 0xf55, 0xc5f, 0xd56, 0xa5a, 0xb53, 0x859, 0x950,
0x7c0, 0x6c9, 0x5c3, 0x4ca, 0x3c6, 0x2cf, 0x1c5, 0xcc ,
0xfcc, 0xec5, 0xdcf, 0xcc6, 0xbca, 0xac3, 0x9c9, 0x8c0,
0x8c0, 0x9c9, 0xac3, 0xbca, 0xcc6, 0xdcf, 0xec5, 0xfcc,
0xcc , 0x1c5, 0x2cf, 0x3c6, 0x4ca, 0x5c3, 0x6c9, 0x7c0,
0x950, 0x859, 0xb53, 0xa5a, 0xd56, 0xc5f, 0xf55, 0xe5c,
0x15c, 0x55 , 0x35f, 0x256, 0x55a, 0x453, 0x759, 0x650,
0xaf0, 0xbf9, 0x8f3, 0x9fa, 0xef6, 0xfff, 0xcf5, 0xdfc,
0x2fc, 0x3f5, 0xff , 0x1f6, 0x6fa, 0x7f3, 0x4f9, 0x5f0,
0xb60, 0xa69, 0x963, 0x86a, 0xf66, 0xe6f, 0xd65, 0xc6c,
0x36c, 0x265, 0x16f, 0x66 , 0x76a, 0x663, 0x569, 0x460,
0xca0, 0xda9, 0xea3, 0xfaa, 0x8a6, 0x9af, 0xaa5, 0xbac,
0x4ac, 0x5a5, 0x6af, 0x7a6, 0xaa , 0x1a3, 0x2a9, 0x3a0,
0xd30, 0xc39, 0xf33, 0xe3a, 0x936, 0x83f, 0xb35, 0xa3c,
0x53c, 0x435, 0x73f, 0x636, 0x13a, 0x33 , 0x339, 0x230,
0xe90, 0xf99, 0xc93, 0xd9a, 0xa96, 0xb9f, 0x895, 0x99c,
0x69c, 0x795, 0x49f, 0x596, 0x29a, 0x393, 0x99 , 0x190,
0xf00, 0xe09, 0xd03, 0xc0a, 0xb06, 0xa0f, 0x905, 0x80c,
0x70c, 0x605, 0x50f, 0x406, 0x30a, 0x203, 0x109, 0x0   };
const int triTable[256][16] =
{{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 8, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 1, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 8, 3, 9, 8, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 8, 3, 1, 2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{9, 2, 10, 0, 2, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{2, 8, 3, 2, 10, 8, 10, 9, 8, -1, -1, -1, -1, -1, -1, -1},
{3, 11, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 11, 2, 8, 11, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 9, 0, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 11, 2, 1, 9, 11, 9, 8, 11, -1, -1, -1, -1, -1, -1, -1},
{3, 10, 1, 11, 10, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 10, 1, 0, 8, 10, 8, 11, 10, -1, -1, -1, -1, -1, -1, -1},
{3, 9, 0, 3, 11, 9, 11, 10, 9, -1, -1, -1, -1, -1, -1, -1},
{9, 8, 10, 10, 8, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{4, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{4, 3, 0, 7, 3, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 1, 9, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{4, 1, 9, 4, 7, 1, 7, 3, 1, -1, -1, -1, -1, -1, -1, -1},
{1, 2, 10, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{3, 4, 7, 3, 0, 4, 1, 2, 10, -1, -1, -1, -1, -1, -1, -1},
{9, 2, 10, 9, 0, 2, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1},
{2, 10, 9, 2, 9, 7, 2, 7, 3, 7, 9, 4, -1, -1, -1, -1},
{8, 4, 7, 3, 11, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{11, 4, 7, 11, 2, 4, 2, 0, 4, -1, -1, -1, -1, -1, -1, -1},
{9, 0, 1, 8, 4, 7, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1},
{4, 7, 11, 9, 4, 11, 9, 11, 2, 9, 2, 1, -1, -1, -1, -1},
{3, 10, 1, 3, 11, 10, 7, 8, 4, -1, -1, -1, -1, -1, -1, -1},
{1, 11, 10, 1, 4, 11, 1, 0, 4, 7, 11, 4, -1, -1, -1, -1},
{4, 7, 8, 9, 0, 11, 9, 11, 10, 11, 0, 3, -1, -1, -1, -1},
{4, 7, 11, 4, 11, 9, 9, 11, 10, -1, -1, -1, -1, -1, -1, -1},
{9, 5, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{9, 5, 4, 0, 8, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 5, 4, 1, 5, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{8, 5, 4, 8, 3, 5, 3, 1, 5, -1, -1, -1, -1, -1, -1, -1},
{1, 2, 10, 9, 5, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{3, 0, 8, 1, 2, 10, 4, 9, 5, -1, -1, -1, -1, -1, -1, -1},
{5, 2, 10, 5, 4, 2, 4, 0, 2, -1, -1, -1, -1, -1, -1, -1},
{2, 10, 5, 3, 2, 5, 3, 5, 4, 3, 4, 8, -1, -1, -1, -1},
{9, 5, 4, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 11, 2, 0, 8, 11, 4, 9, 5, -1, -1, -1, -1, -1, -1, -1},
{0, 5, 4, 0, 1, 5, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1},
{2, 1, 5, 2, 5, 8, 2, 8, 11, 4, 8, 5, -1, -1, -1, -1},
{10, 3, 11, 10, 1, 3, 9, 5, 4, -1, -1, -1, -1, -1, -1, -1},
{4, 9, 5, 0, 8, 1, 8, 10, 1, 8, 11, 10, -1, -1, -1, -1},
{5, 4, 0, 5, 0, 11, 5, 11, 10, 11, 0, 3, -1, -1, -1, -1},
{5, 4, 8, 5, 8, 10, 10, 8, 11, -1, -1, -1, -1, -1, -1, -1},
{9, 7, 8, 5, 7, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{9, 3, 0, 9, 5, 3, 5, 7, 3, -1, -1, -1, -1, -1, -1, -1},
{0, 7, 8, 0, 1, 7, 1, 5, 7, -1, -1, -1, -1, -1, -1, -1},
{1, 5, 3, 3, 5, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{9, 7, 8, 9, 5, 7, 10, 1, 2, -1, -1, -1, -1, -1, -1, -1},
{10, 1, 2, 9, 5, 0, 5, 3, 0, 5, 7, 3, -1, -1, -1, -1},
{8, 0, 2, 8, 2, 5, 8, 5, 7, 10, 5, 2, -1, -1, -1, -1},
{2, 10, 5, 2, 5, 3, 3, 5, 7, -1, -1, -1, -1, -1, -1, -1},
{7, 9, 5, 7, 8, 9, 3, 11, 2, -1, -1, -1, -1, -1, -1, -1},
{9, 5, 7, 9, 7, 2, 9, 2, 0, 2, 7, 11, -1, -1, -1, -1},
{2, 3, 11, 0, 1, 8, 1, 7, 8, 1, 5, 7, -1, -1, -1, -1},
{11, 2, 1, 11, 1, 7, 7, 1, 5, -1, -1, -1, -1, -1, -1, -1},
{9, 5, 8, 8, 5, 7, 10, 1, 3, 10, 3, 11, -1, -1, -1, -1},
{5, 7, 0, 5, 0, 9, 7, 11, 0, 1, 0, 10, 11, 10, 0, -1},
{11, 10, 0, 11, 0, 3, 10, 5, 0, 8, 0, 7, 5, 7, 0, -1},
{11, 10, 5, 7, 11, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{10, 6, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 8, 3, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{9, 0, 1, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 8, 3, 1, 9, 8, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1},
{1, 6, 5, 2, 6, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 6, 5, 1, 2, 6, 3, 0, 8, -1, -1, -1, -1, -1, -1, -1},
{9, 6, 5, 9, 0, 6, 0, 2, 6, -1, -1, -1, -1, -1, -1, -1},
{5, 9, 8, 5, 8, 2, 5, 2, 6, 3, 2, 8, -1, -1, -1, -1},
{2, 3, 11, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{11, 0, 8, 11, 2, 0, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1},
{0, 1, 9, 2, 3, 11, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1},
{5, 10, 6, 1, 9, 2, 9, 11, 2, 9, 8, 11, -1, -1, -1, -1},
{6, 3, 11, 6, 5, 3, 5, 1, 3, -1, -1, -1, -1, -1, -1, -1},
{0, 8, 11, 0, 11, 5, 0, 5, 1, 5, 11, 6, -1, -1, -1, -1},
{3, 11, 6, 0, 3, 6, 0, 6, 5, 0, 5, 9, -1, -1, -1, -1},
{6, 5, 9, 6, 9, 11, 11, 9, 8, -1, -1, -1, -1, -1, -1, -1},
{5, 10, 6, 4, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{4, 3, 0, 4, 7, 3, 6, 5, 10, -1, -1, -1, -1, -1, -1, -1},
{1, 9, 0, 5, 10, 6, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1},
{10, 6, 5, 1, 9, 7, 1, 7, 3, 7, 9, 4, -1, -1, -1, -1},
{6, 1, 2, 6, 5, 1, 4, 7, 8, -1, -1, -1, -1, -1, -1, -1},
{1, 2, 5, 5, 2, 6, 3, 0, 4, 3, 4, 7, -1, -1, -1, -1},
{8, 4, 7, 9, 0, 5, 0, 6, 5, 0, 2, 6, -1, -1, -1, -1},
{7, 3, 9, 7, 9, 4, 3, 2, 9, 5, 9, 6, 2, 6, 9, -1},
{3, 11, 2, 7, 8, 4, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1},
{5, 10, 6, 4, 7, 2, 4, 2, 0, 2, 7, 11, -1, -1, -1, -1},
{0, 1, 9, 4, 7, 8, 2, 3, 11, 5, 10, 6, -1, -1, -1, -1},
{9, 2, 1, 9, 11, 2, 9, 4, 11, 7, 11, 4, 5, 10, 6, -1},
{8, 4, 7, 3, 11, 5, 3, 5, 1, 5, 11, 6, -1, -1, -1, -1},
{5, 1, 11, 5, 11, 6, 1, 0, 11, 7, 11, 4, 0, 4, 11, -1},
{0, 5, 9, 0, 6, 5, 0, 3, 6, 11, 6, 3, 8, 4, 7, -1},
{6, 5, 9, 6, 9, 11, 4, 7, 9, 7, 11, 9, -1, -1, -1, -1},
{10, 4, 9, 6, 4, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{4, 10, 6, 4, 9, 10, 0, 8, 3, -1, -1, -1, -1, -1, -1, -1},
{10, 0, 1, 10, 6, 0, 6, 4, 0, -1, -1, -1, -1, -1, -1, -1},
{8, 3, 1, 8, 1, 6, 8, 6, 4, 6, 1, 10, -1, -1, -1, -1},
{1, 4, 9, 1, 2, 4, 2, 6, 4, -1, -1, -1, -1, -1, -1, -1},
{3, 0, 8, 1, 2, 9, 2, 4, 9, 2, 6, 4, -1, -1, -1, -1},
{0, 2, 4, 4, 2, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{8, 3, 2, 8, 2, 4, 4, 2, 6, -1, -1, -1, -1, -1, -1, -1},
{10, 4, 9, 10, 6, 4, 11, 2, 3, -1, -1, -1, -1, -1, -1, -1},
{0, 8, 2, 2, 8, 11, 4, 9, 10, 4, 10, 6, -1, -1, -1, -1},
{3, 11, 2, 0, 1, 6, 0, 6, 4, 6, 1, 10, -1, -1, -1, -1},
{6, 4, 1, 6, 1, 10, 4, 8, 1, 2, 1, 11, 8, 11, 1, -1},
{9, 6, 4, 9, 3, 6, 9, 1, 3, 11, 6, 3, -1, -1, -1, -1},
{8, 11, 1, 8, 1, 0, 11, 6, 1, 9, 1, 4, 6, 4, 1, -1},
{3, 11, 6, 3, 6, 0, 0, 6, 4, -1, -1, -1, -1, -1, -1, -1},
{6, 4, 8, 11, 6, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{7, 10, 6, 7, 8, 10, 8, 9, 10, -1, -1, -1, -1, -1, -1, -1},
{0, 7, 3, 0, 10, 7, 0, 9, 10, 6, 7, 10, -1, -1, -1, -1},
{10, 6, 7, 1, 10, 7, 1, 7, 8, 1, 8, 0, -1, -1, -1, -1},
{10, 6, 7, 10, 7, 1, 1, 7, 3, -1, -1, -1, -1, -1, -1, -1},
{1, 2, 6, 1, 6, 8, 1, 8, 9, 8, 6, 7, -1, -1, -1, -1},
{2, 6, 9, 2, 9, 1, 6, 7, 9, 0, 9, 3, 7, 3, 9, -1},
{7, 8, 0, 7, 0, 6, 6, 0, 2, -1, -1, -1, -1, -1, -1, -1},
{7, 3, 2, 6, 7, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{2, 3, 11, 10, 6, 8, 10, 8, 9, 8, 6, 7, -1, -1, -1, -1},
{2, 0, 7, 2, 7, 11, 0, 9, 7, 6, 7, 10, 9, 10, 7, -1},
{1, 8, 0, 1, 7, 8, 1, 10, 7, 6, 7, 10, 2, 3, 11, -1},
{11, 2, 1, 11, 1, 7, 10, 6, 1, 6, 7, 1, -1, -1, -1, -1},
{8, 9, 6, 8, 6, 7, 9, 1, 6, 11, 6, 3, 1, 3, 6, -1},
{0, 9, 1, 11, 6, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{7, 8, 0, 7, 0, 6, 3, 11, 0, 11, 6, 0, -1, -1, -1, -1},
{7, 11, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{7, 6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{3, 0, 8, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 1, 9, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{8, 1, 9, 8, 3, 1, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1},
{10, 1, 2, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 2, 10, 3, 0, 8, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1},
{2, 9, 0, 2, 10, 9, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1},
{6, 11, 7, 2, 10, 3, 10, 8, 3, 10, 9, 8, -1, -1, -1, -1},
{7, 2, 3, 6, 2, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{7, 0, 8, 7, 6, 0, 6, 2, 0, -1, -1, -1, -1, -1, -1, -1},
{2, 7, 6, 2, 3, 7, 0, 1, 9, -1, -1, -1, -1, -1, -1, -1},
{1, 6, 2, 1, 8, 6, 1, 9, 8, 8, 7, 6, -1, -1, -1, -1},
{10, 7, 6, 10, 1, 7, 1, 3, 7, -1, -1, -1, -1, -1, -1, -1},
{10, 7, 6, 1, 7, 10, 1, 8, 7, 1, 0, 8, -1, -1, -1, -1},
{0, 3, 7, 0, 7, 10, 0, 10, 9, 6, 10, 7, -1, -1, -1, -1},
{7, 6, 10, 7, 10, 8, 8, 10, 9, -1, -1, -1, -1, -1, -1, -1},
{6, 8, 4, 11, 8, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{3, 6, 11, 3, 0, 6, 0, 4, 6, -1, -1, -1, -1, -1, -1, -1},
{8, 6, 11, 8, 4, 6, 9, 0, 1, -1, -1, -1, -1, -1, -1, -1},
{9, 4, 6, 9, 6, 3, 9, 3, 1, 11, 3, 6, -1, -1, -1, -1},
{6, 8, 4, 6, 11, 8, 2, 10, 1, -1, -1, -1, -1, -1, -1, -1},
{1, 2, 10, 3, 0, 11, 0, 6, 11, 0, 4, 6, -1, -1, -1, -1},
{4, 11, 8, 4, 6, 11, 0, 2, 9, 2, 10, 9, -1, -1, -1, -1},
{10, 9, 3, 10, 3, 2, 9, 4, 3, 11, 3, 6, 4, 6, 3, -1},
{8, 2, 3, 8, 4, 2, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1},
{0, 4, 2, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 9, 0, 2, 3, 4, 2, 4, 6, 4, 3, 8, -1, -1, -1, -1},
{1, 9, 4, 1, 4, 2, 2, 4, 6, -1, -1, -1, -1, -1, -1, -1},
{8, 1, 3, 8, 6, 1, 8, 4, 6, 6, 10, 1, -1, -1, -1, -1},
{10, 1, 0, 10, 0, 6, 6, 0, 4, -1, -1, -1, -1, -1, -1, -1},
{4, 6, 3, 4, 3, 8, 6, 10, 3, 0, 3, 9, 10, 9, 3, -1},
{10, 9, 4, 6, 10, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{4, 9, 5, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 8, 3, 4, 9, 5, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1},
{5, 0, 1, 5, 4, 0, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1},
{11, 7, 6, 8, 3, 4, 3, 5, 4, 3, 1, 5, -1, -1, -1, -1},
{9, 5, 4, 10, 1, 2, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1},
{6, 11, 7, 1, 2, 10, 0, 8, 3, 4, 9, 5, -1, -1, -1, -1},
{7, 6, 11, 5, 4, 10, 4, 2, 10, 4, 0, 2, -1, -1, -1, -1},
{3, 4, 8, 3, 5, 4, 3, 2, 5, 10, 5, 2, 11, 7, 6, -1},
{7, 2, 3, 7, 6, 2, 5, 4, 9, -1, -1, -1, -1, -1, -1, -1},
{9, 5, 4, 0, 8, 6, 0, 6, 2, 6, 8, 7, -1, -1, -1, -1},
{3, 6, 2, 3, 7, 6, 1, 5, 0, 5, 4, 0, -1, -1, -1, -1},
{6, 2, 8, 6, 8, 7, 2, 1, 8, 4, 8, 5, 1, 5, 8, -1},
{9, 5, 4, 10, 1, 6, 1, 7, 6, 1, 3, 7, -1, -1, -1, -1},
{1, 6, 10, 1, 7, 6, 1, 0, 7, 8, 7, 0, 9, 5, 4, -1},
{4, 0, 10, 4, 10, 5, 0, 3, 10, 6, 10, 7, 3, 7, 10, -1},
{7, 6, 10, 7, 10, 8, 5, 4, 10, 4, 8, 10, -1, -1, -1, -1},
{6, 9, 5, 6, 11, 9, 11, 8, 9, -1, -1, -1, -1, -1, -1, -1},
{3, 6, 11, 0, 6, 3, 0, 5, 6, 0, 9, 5, -1, -1, -1, -1},
{0, 11, 8, 0, 5, 11, 0, 1, 5, 5, 6, 11, -1, -1, -1, -1},
{6, 11, 3, 6, 3, 5, 5, 3, 1, -1, -1, -1, -1, -1, -1, -1},
{1, 2, 10, 9, 5, 11, 9, 11, 8, 11, 5, 6, -1, -1, -1, -1},
{0, 11, 3, 0, 6, 11, 0, 9, 6, 5, 6, 9, 1, 2, 10, -1},
{11, 8, 5, 11, 5, 6, 8, 0, 5, 10, 5, 2, 0, 2, 5, -1},
{6, 11, 3, 6, 3, 5, 2, 10, 3, 10, 5, 3, -1, -1, -1, -1},
{5, 8, 9, 5, 2, 8, 5, 6, 2, 3, 8, 2, -1, -1, -1, -1},
{9, 5, 6, 9, 6, 0, 0, 6, 2, -1, -1, -1, -1, -1, -1, -1},
{1, 5, 8, 1, 8, 0, 5, 6, 8, 3, 8, 2, 6, 2, 8, -1},
{1, 5, 6, 2, 1, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 3, 6, 1, 6, 10, 3, 8, 6, 5, 6, 9, 8, 9, 6, -1},
{10, 1, 0, 10, 0, 6, 9, 5, 0, 5, 6, 0, -1, -1, -1, -1},
{0, 3, 8, 5, 6, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{10, 5, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{11, 5, 10, 7, 5, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{11, 5, 10, 11, 7, 5, 8, 3, 0, -1, -1, -1, -1, -1, -1, -1},
{5, 11, 7, 5, 10, 11, 1, 9, 0, -1, -1, -1, -1, -1, -1, -1},
{10, 7, 5, 10, 11, 7, 9, 8, 1, 8, 3, 1, -1, -1, -1, -1},
{11, 1, 2, 11, 7, 1, 7, 5, 1, -1, -1, -1, -1, -1, -1, -1},
{0, 8, 3, 1, 2, 7, 1, 7, 5, 7, 2, 11, -1, -1, -1, -1},
{9, 7, 5, 9, 2, 7, 9, 0, 2, 2, 11, 7, -1, -1, -1, -1},
{7, 5, 2, 7, 2, 11, 5, 9, 2, 3, 2, 8, 9, 8, 2, -1},
{2, 5, 10, 2, 3, 5, 3, 7, 5, -1, -1, -1, -1, -1, -1, -1},
{8, 2, 0, 8, 5, 2, 8, 7, 5, 10, 2, 5, -1, -1, -1, -1},
{9, 0, 1, 5, 10, 3, 5, 3, 7, 3, 10, 2, -1, -1, -1, -1},
{9, 8, 2, 9, 2, 1, 8, 7, 2, 10, 2, 5, 7, 5, 2, -1},
{1, 3, 5, 3, 7, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 8, 7, 0, 7, 1, 1, 7, 5, -1, -1, -1, -1, -1, -1, -1},
{9, 0, 3, 9, 3, 5, 5, 3, 7, -1, -1, -1, -1, -1, -1, -1},
{9, 8, 7, 5, 9, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{5, 8, 4, 5, 10, 8, 10, 11, 8, -1, -1, -1, -1, -1, -1, -1},
{5, 0, 4, 5, 11, 0, 5, 10, 11, 11, 3, 0, -1, -1, -1, -1},
{0, 1, 9, 8, 4, 10, 8, 10, 11, 10, 4, 5, -1, -1, -1, -1},
{10, 11, 4, 10, 4, 5, 11, 3, 4, 9, 4, 1, 3, 1, 4, -1},
{2, 5, 1, 2, 8, 5, 2, 11, 8, 4, 5, 8, -1, -1, -1, -1},
{0, 4, 11, 0, 11, 3, 4, 5, 11, 2, 11, 1, 5, 1, 11, -1},
{0, 2, 5, 0, 5, 9, 2, 11, 5, 4, 5, 8, 11, 8, 5, -1},
{9, 4, 5, 2, 11, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{2, 5, 10, 3, 5, 2, 3, 4, 5, 3, 8, 4, -1, -1, -1, -1},
{5, 10, 2, 5, 2, 4, 4, 2, 0, -1, -1, -1, -1, -1, -1, -1},
{3, 10, 2, 3, 5, 10, 3, 8, 5, 4, 5, 8, 0, 1, 9, -1},
{5, 10, 2, 5, 2, 4, 1, 9, 2, 9, 4, 2, -1, -1, -1, -1},
{8, 4, 5, 8, 5, 3, 3, 5, 1, -1, -1, -1, -1, -1, -1, -1},
{0, 4, 5, 1, 0, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{8, 4, 5, 8, 5, 3, 9, 0, 5, 0, 3, 5, -1, -1, -1, -1},
{9, 4, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{4, 11, 7, 4, 9, 11, 9, 10, 11, -1, -1, -1, -1, -1, -1, -1},
{0, 8, 3, 4, 9, 7, 9, 11, 7, 9, 10, 11, -1, -1, -1, -1},
{1, 10, 11, 1, 11, 4, 1, 4, 0, 7, 4, 11, -1, -1, -1, -1},
{3, 1, 4, 3, 4, 8, 1, 10, 4, 7, 4, 11, 10, 11, 4, -1},
{4, 11, 7, 9, 11, 4, 9, 2, 11, 9, 1, 2, -1, -1, -1, -1},
{9, 7, 4, 9, 11, 7, 9, 1, 11, 2, 11, 1, 0, 8, 3, -1},
{11, 7, 4, 11, 4, 2, 2, 4, 0, -1, -1, -1, -1, -1, -1, -1},
{11, 7, 4, 11, 4, 2, 8, 3, 4, 3, 2, 4, -1, -1, -1, -1},
{2, 9, 10, 2, 7, 9, 2, 3, 7, 7, 4, 9, -1, -1, -1, -1},
{9, 10, 7, 9, 7, 4, 10, 2, 7, 8, 7, 0, 2, 0, 7, -1},
{3, 7, 10, 3, 10, 2, 7, 4, 10, 1, 10, 0, 4, 0, 10, -1},
{1, 10, 2, 8, 7, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{4, 9, 1, 4, 1, 7, 7, 1, 3, -1, -1, -1, -1, -1, -1, -1},
{4, 9, 1, 4, 1, 7, 0, 8, 1, 8, 7, 1, -1, -1, -1, -1},
{4, 0, 3, 7, 4, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{4, 8, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{9, 10, 8, 10, 11, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{3, 0, 9, 3, 9, 11, 11, 9, 10, -1, -1, -1, -1, -1, -1, -1},
{0, 1, 10, 0, 10, 8, 8, 10, 11, -1, -1, -1, -1, -1, -1, -1},
{3, 1, 10, 11, 3, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 2, 11, 1, 11, 9, 9, 11, 8, -1, -1, -1, -1, -1, -1, -1},
{3, 0, 9, 3, 9, 11, 1, 2, 9, 2, 11, 9, -1, -1, -1, -1},
{0, 2, 11, 8, 0, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{3, 2, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{2, 3, 8, 2, 8, 10, 10, 8, 9, -1, -1, -1, -1, -1, -1, -1},
{9, 10, 2, 0, 9, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{2, 3, 8, 2, 8, 10, 0, 1, 8, 1, 10, 8, -1, -1, -1, -1},
{1, 10, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 3, 8, 9, 1, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 9, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 3, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}};
