// *****************************
// ***    R e n d e r e r    ***
// *****************************
// ***    by Fuzzware 2001   ***
// *****************************
// 
// Engine 3D che raggruppa tutte
// le funzioni implementate

#include "renderer.h"

bool isSubString(char *src,const char *search) {
	int pos;
	int maxpos=strlen(search)-1;
	int len=strlen(src);
	for(int i=0;i<len;i++) {
		if ((i==0)||((i>1)&&src[i-1]==' ')) {
			pos=0;
			while (src[i]!=' ') {
				if (src[i]==search[pos]) pos++;
				if ((pos>maxpos)&&src[i+1]==' ') return true;
				i++;
			}
		}
	}
	return false;
}

// *****************************
// *****************************
// ***       T i m e r       ***
// *****************************
// *****************************

void Timer::init() {
	message("Query performance timer: ");
	if (!QueryPerformanceFrequency((LARGE_INTEGER*)&freq)) {
		perfTimer=FALSE;
		resolution=1.0f/1000.0f;
		freq=1000;
		message("not found, using normal timer.\n");
	}
	else {
		perfTimer=TRUE;
		resolution=(float)(((double)1.0)/((double)freq));
		message("found, using performance timer.\n");
	}
	setTime();
}

void Timer::setTime() {
	__int64 t;
	if (perfTimer) QueryPerformanceCounter((LARGE_INTEGER*)&t);
	else t=timeGetTime();
	time=(float(t)*resolution)*1000.0f;
}

bool Timer::lessTime(float ms) {
	__int64 t;
	if (perfTimer) QueryPerformanceCounter((LARGE_INTEGER*)&t);
	else t=timeGetTime();
	float time2=(float(t)*resolution)*1000.0f;
	if (time2-time<ms) return true;
	else return false;
}

float Timer::getTime() {
	__int64 t;
	if (perfTimer) QueryPerformanceCounter((LARGE_INTEGER*)&t);
	else t=timeGetTime();
	float time2=(float(t)*resolution)*1000.0f;
	return (time2-time);
}

// *****************************
// *****************************
// ***    R e n d e r e r    ***
// *****************************
// *****************************

Renderer::Renderer() {
	_control87(MCW_EM,MCW_EM);
	devHandle=NULL;
	ra=ga=ba=0.0f;
	x01=y01=wscr1=hscr1=0;
	x02=y02=wscr2=hscr2=0;
	view1=FRONT;
	view2=FRONT;
	maxFPS=50.0f;
	visFPS=false;
	mainLoopCall=0;
	draw2DCall=0;
	//display=0;
}

Renderer::~Renderer() {
	unInit();
}

bool Renderer::init(HDC hdc) {
	if (hdc==NULL) {
		message("Null device handle, init failed.\n");
		return false;
	}
	devHandle=hdc;
	int vport[4];
	glGetIntegerv(GL_VIEWPORT,vport);
	x01=vport[0];y01=vport[1];
	wscr1=vport[2];hscr1=vport[3];
	// Check for multitexturing
	char *extensions;
	extensions=strdup((char*)glGetString(GL_EXTENSIONS));
	if (isSubString(extensions,"GL_ARB_multitexture")&&isSubString(extensions,"GL_EXT_texture_env_add")) {
		glGetIntegerv(GL_MAX_TEXTURE_UNITS_ARB,&texUnits);
		glMultiTexCoord1fARB=(PFNGLMULTITEXCOORD1FARBPROC)wglGetProcAddress("glMultiTexCoord1fARB");
		glMultiTexCoord2fARB=(PFNGLMULTITEXCOORD2FARBPROC)wglGetProcAddress("glMultiTexCoord2fARB");
		glMultiTexCoord3fARB=(PFNGLMULTITEXCOORD3FARBPROC)wglGetProcAddress("glMultiTexCoord3fARB");
		glMultiTexCoord4fARB=(PFNGLMULTITEXCOORD4FARBPROC)wglGetProcAddress("glMultiTexCoord4fARB");
		glActiveTextureARB=(PFNGLACTIVETEXTUREARBPROC)wglGetProcAddress("glActiveTextureARB");
		glClientActiveTextureARB=(PFNGLCLIENTACTIVETEXTUREARBPROC)wglGetProcAddress("glClientActiveTextureARB");
		multiTex=true;
		message("Requested multitexturing features supported: %d texture units, GL_ADD texenv.\n",texUnits);
	}
	else {
		multiTex=false;
		message("Requested multitexturing features not supported.\n");
	}

	// OpenGL settings
	glShadeModel(GL_SMOOTH);
	glClearColor(0.0,0.0,0.0,1.0);
	glClearDepth(1.0);

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glEnable(GL_CULL_FACE);
	glEnable(GL_BLEND);
	glEnable(GL_TEXTURE_2D);
	glDisable(GL_TEXTURE_GEN_S);
	glDisable(GL_TEXTURE_GEN_T);
	glDisable(GL_POLYGON_SMOOTH);
	glDisable(GL_LINE_SMOOTH);
	glDisable(GL_POINT_SMOOTH);
	glDisable(GL_COLOR_MATERIAL);
	glDepthFunc(GL_LEQUAL);
	glDepthMask(GL_TRUE);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glHint(GL_FOG_HINT,GL_NICEST);
	glHint(GL_POLYGON_SMOOTH_HINT,GL_NICEST);
	glHint(GL_LINE_SMOOTH_HINT,GL_NICEST);
	glHint(GL_POINT_SMOOTH_HINT,GL_NICEST);
	glCullFace(GL_BACK);
	glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
	glTexGeni(GL_S,GL_TEXTURE_GEN_MODE,GL_SPHERE_MAP);
	glTexGeni(GL_T,GL_TEXTURE_GEN_MODE,GL_SPHERE_MAP);	
	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,0);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_FALSE);
	GLfloat amb[]={ra,ga,ba,1.0};
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT,amb);
	
	fnt.buildBitmapFont(devHandle,"Arial",16);
	
   //cam.smooth(0.44);
	glMatrixMode(GL_TEXTURE);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	timer.init();
	
	return true;
}

void Renderer::unInit() {
	if (devHandle==NULL) return;
	fnt.kill();
	scene.kill();
	devHandle=NULL;
}

void Renderer::addObject(Object *obj) {
	scene.addObject(obj);
}

int Renderer::mainLoop() {
	MSG msg;
	int counter=0;
	SYSTEMTIME sysTime;
	FILETIME fTime;
	unsigned int start,stop;
	float fps=0;
	GetSystemTime(&sysTime);
	SystemTimeToFileTime(&sysTime,&fTime);
	start=fTime.dwLowDateTime;
	bool quit=false;
	float lastFps=0;
	Vector lp(0.0,0.0,-20.0);

	while(!quit) {
		if (PeekMessage(&msg,NULL,0,0,PM_REMOVE))
			if (msg.message==WM_QUIT) quit=TRUE;
			else {
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
		else {
			if ((++counter)>=10) {
				GetSystemTime(&sysTime);
				SystemTimeToFileTime(&sysTime,&fTime);
				stop=fTime.dwLowDateTime;
				if (stop==start) fps=1000.0;
				else fps=(10)*10000000/((float)(stop-start));
				start=stop;
				fps=0.5f*fps+0.5f*lastFps;
				lastFps=fps;
				counter=0;
			}
			
			timer.setTime();
			
			if (mainLoopCall!=0) (*mainLoopCall)();
			
			glViewport(x01,y01,wscr1,hscr1);
			glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
			GLfloat amb[]={ra,ga,ba,1.0};
			glLightModelfv(GL_LIGHT_MODEL_AMBIENT,amb);
						
			scene.render(&cam.getView(view1));

			// Picture in Picture
			if (wscr2!=0) {
				glEnable(GL_SCISSOR_TEST);
				glScissor(x02-1,y02-1,wscr2+2,hscr2+2);
				glClearColor(0.5,0.5,0.5,1.0);
				glClear(GL_COLOR_BUFFER_BIT);
				glScissor(x02,y02,wscr2,hscr2);
				glClearColor(0.0,0.0,0.0,1.0);
				glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
				glDisable(GL_SCISSOR_TEST);
				glViewport(x02,y02,wscr2,hscr2);
				scene.render(&cam.getView(view2));
				glViewport(x01,y01,wscr1,hscr1);
			}


			//scritte!!!	
			glLoadIdentity();		
			glDisable(GL_LIGHTING);
		   glEnable(GL_COLOR_MATERIAL);
		   glDisable(GL_DEPTH_TEST);
			glDisable(GL_BLEND);
		   glDisable(GL_TEXTURE_2D);
			glMatrixMode(GL_PROJECTION);
			glPushMatrix();
			glLoadIdentity();
			glOrtho(0.0,800,0.0,600,-1.0,1.0);
			glColor3f(1.0,0.0,0.0);
			if (visFPS) fnt.print(650,570,"FPS: %.2f",fps);
			
			if (draw2DCall!=0) (*draw2DCall)();
			if (af!=0.0) {
				glColor4f(rf,gf,bf,af);
				glEnable(GL_BLEND);
				glBegin(GL_QUADS);
					glVertex2f(0.0,0.0);
					glVertex2f(800.0,0.0);
					glVertex2f(800.0,600.0);
					glVertex2f(0.0,600.0);
				glEnd();
			}
			
			
			glEnable(GL_LIGHTING);
		   glDisable(GL_COLOR_MATERIAL);
		   glEnable(GL_DEPTH_TEST);
			glEnable(GL_BLEND);
			glPopMatrix();
			glMatrixMode(GL_MODELVIEW);

			SwapBuffers(devHandle);
			
			while(timer.lessTime(1000.0f/maxFPS));
		}
	}
	return msg.wParam;
}

void Renderer::setBackground(Background *bg) {
	scene.setBackground(bg);
}

void Renderer::setAmbient(float r,float g,float b) {
	ra=r;ga=g;ba=b;
}

void Renderer::setFadeQuad(float r,float g,float b,float a) {
	rf=r;gf=g;bf=b;af=a;
}

void Renderer::setPerspective(float fov,float zn,float zf) {
	scene.setPerspective(fov,zn,zf);
}

void Renderer::setPrimaryView(int xx0,int yy0,int w,int h,int tp) {
	if ((w==0)||(h==0)) return;
	else {
		x01=xx0;y01=yy0;
		wscr1=w;hscr1=h;
		view1=tp;
	}
}

void Renderer::setSecondaryView(int xx0,int yy0,int w,int h,int tp) {
	if ((w==0)||(h==0)) x02=y02=wscr2=hscr2=view2=0;
	else {
		x02=xx0;y02=yy0;
		wscr2=w;hscr2=h;
		view2=tp;
	}
}

void Renderer::setMaxFramerate(float mf,bool b) {
	if (mf>5.0f) maxFPS=mf;
	visFPS=b;
}

void Renderer::setDepthFog(bool en,float dist,float rf,float gf,float bf) {
	//dist: distanza alla quale si ha 1% del colore originale
	float density=sqrt(4.60517/(dist*dist));
	if (!en) {
		glDisable(GL_FOG);
		glClearColor(0.0,0.0,0.0,1.0);
		return;
	}
	GLfloat fogCol[4]={rf,gf,bf,1.0};
	glFogi(GL_FOG_MODE,GL_EXP2);
	//glFogi(GL_FOG_MODE,GL_LINEAR);
	//glFogf(GL_FOG_START,20.0);
	//glFogf(GL_FOG_END,100.0);//zfar o meno, no?
	glFogfv(GL_FOG_COLOR,fogCol);
	glFogf(GL_FOG_DENSITY,density);
	glEnable(GL_FOG);
	glClearColor(rf,gf,bf,1.0);
}

void Renderer::setCallBack(int n,void (*func)()) {
	switch (n) {
		case MAINLOOP:
			mainLoopCall=func;
			break;
		case DRAW2D:
			draw2DCall=func;
			break;
	}
}

Font& Renderer::getFont() {
	return fnt;	
}

Camera& Renderer::getCamera() {
	return cam;
}

