// *****************************
// ***    T e x t u r e s    ***
// *****************************
// ***    by Fuzzware 2001   ***
// *****************************
// 
// Caricamento di file bmp e
// uso come texture

#include "textures.h"

TexManager tMan;

Texture::Texture() {
	ready=false;
}

Texture::~Texture() {
	if (ready) kill();
}

bool Texture::setDiffuseMap(char *fname,int detail) {
	if (ready) return false;
	name=tMan.bind(fname,detail,NORMAL);
	if (name==-1) return false;
	ready=true;
	return true;
}

bool Texture::setBumpMap(char *fname,int detail,int pass) {
	if (ready) return false;
	name=tMan.bind(fname,detail,pass);
	if (name==-1) return false;
	ready=true;
	return true;
}

void Texture::apply() {
	if (!ready) {
		glDisable(GL_TEXTURE_2D);
		return;
	}
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D,name);
}

void Texture::kill() {
	if (!ready) return;
	tMan.kill(name);
	ready=false;
}


TexManager::TexManager() {
	allTex=0;
}

TexManager::~TexManager() {
	if (allTex==0) return;
	texFile *next=allTex;
	texFile *tmp;
	while(next!=0) {
		tmp=next;
		next=next->next;
		delete tmp;
	}
}

GLuint TexManager::bind(char *fname,int detail,int tp) {
	texFile *curr=allTex;
	texFile *prev;
	message("Loading texture file: %s\n",fname);
	while(curr!=0) {
		if ((strcmpi(fname,curr->fname)==0)&&(detail==curr->detail)&&(tp==curr->tp)) {
			message("Texture already loaded, binding to existing name.\n");
			curr->num++;
			return curr->bind;
		}
		prev=curr;
		curr=curr->next;
	}
	GLuint tmp;
	// Non trovata, va caricata
	Image img;
	GLint magFilter,minFilter,minFilterMip;
	if (!img.loadBMP(fname)) {
		message("Error: could not open file!\n\n");
		return -1;
	}
	int w=img.getWidth();
	int h=img.getHeight();
	message("Loaded %dx%d image.\n",w,h);
	if ((detail==DETAIL1)||(detail==DETAIL2)) {
		magFilter=minFilter=GL_LINEAR;
		minFilterMip=GL_LINEAR_MIPMAP_NEAREST;
	}
	else {
		magFilter=minFilter=GL_NEAREST;
		minFilterMip=GL_NEAREST_MIPMAP_NEAREST;
	}
	if (tp==BUMP1) {
		img.scale(0.5);
	}	
	if (tp==BUMP2) {
		img.scale(2.0);
		img.invert();
		img.scale(0.5);
	}	
	glGenTextures(1,&tmp);
	glBindTexture(GL_TEXTURE_2D,tmp);
	glPixelStorei(GL_UNPACK_ALIGNMENT,1);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
	if (detail==DETAIL2)	{
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,magFilter);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,minFilterMip);
		gluBuild2DMipmaps(GL_TEXTURE_2D,GL_RGB8,w,h,GL_RGB,GL_UNSIGNED_BYTE,img.getArray());
	}
	else {
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,magFilter);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,minFilter);
		glTexImage2D(GL_TEXTURE_2D,0,GL_RGB8,w,h,0,GL_RGB,GL_UNSIGNED_BYTE,img.getArray());
	}
	glFlush();
	if (allTex==0) {
		allTex=new texFile();
		curr=allTex;
	}
	else {
		prev->next=new texFile();
		curr=prev->next;
	}
	strcpy(curr->fname,fname);
	curr->bind=tmp;
	curr->num=1;
	curr->detail=detail;
	curr->tp=tp;
	curr->next=0;
	return curr->bind;
}

void TexManager::kill(unsigned int bind) {
	texFile *curr=allTex;
	texFile *prev=0;
	while(curr!=0) {
		if (bind==curr->bind) {
			curr->num--;
			if (curr->num==0) {
				glDeleteTextures(1,&(curr->bind));
				if (prev==0) allTex=curr->next;
				else prev->next=curr->next;
				message("Deleted texture %s\n",curr->fname);
				delete curr;
				return;
			}
			return;
		}
		prev=curr;
		curr=curr->next;
	}
}
